import typing

from commercetools import types
from commercetools._schemas._product import (
    ProductProjectionPagedQueryResponseSchema,
    ProductProjectionSchema,
    ProductSchema,
)
from commercetools.services.product_projections import _ProductProjectionQuerySchema
from commercetools.testing import utils
from commercetools.testing.abstract import ServiceBackend
from commercetools.testing.utils import create_commercetools_response


class ProductProjectionsBackend(ServiceBackend):
    service_path = "product-projections"
    _schema_query_params = _ProductProjectionQuerySchema

    def urls(self):
        return [
            ("^$", "GET", self.query),
            ("^search", "POST", self.search),
            ("^key=(?P<key>[^/]+)$", "GET", self.get_by_key),
            ("^(?P<id>[^/]+)$", "GET", self.get_by_id),
        ]

    def query(self, request):
        params = utils.parse_request_params(_ProductProjectionQuerySchema, request)
        limit = params.get("limit")

        results = []
        found = 0
        for obj in self.model.objects.values():
            expanded_obj = self._expand(request, obj)
            product = ProductSchema().load(expanded_obj)

            result = self._convert_product_projection(product, params["staged"])
            if result:
                results.append(result)
                found += 1

            if limit is not None and found == limit:
                break

        data = {
            "count": len(results),
            "total": len(results),
            "offset": 0,
            "results": results,
            "limit": limit,
        }
        content = ProductProjectionPagedQueryResponseSchema().dumps(data)
        return create_commercetools_response(request, text=content)

    def search(self, request):
        params = utils.parse_request_params(_ProductProjectionQuerySchema, request)

        limit = params.get("limit")

        results = []
        found = 0
        for obj in self.model.objects.values():
            expanded_obj = self._expand(request, obj)
            product = ProductSchema().load(expanded_obj)

            result = self._convert_product_projection(product, params["staged"])
            if result:
                results.append(results)
                found += 1

            if limit is not None and found == limit:
                break

        data = {
            "count": len(results),
            "total": len(results),  # FIXME
            "offset": 0,
            "results": results,
            "limit": limit,
        }
        content = ProductProjectionPagedQueryResponseSchema().dumps(data)
        return create_commercetools_response(request, text=content)

    def get_by_id(self, request, id):
        params = utils.parse_request_params(_ProductProjectionQuerySchema, request)
        obj = self.model.get_by_id(id)
        if obj:
            expanded_obj = self._expand(request, obj)
            product = ProductSchema().load(expanded_obj)

            result = self._convert_product_projection(product, params["staged"])
            if not result:
                return create_commercetools_response(request, status_code=404)
            content = ProductProjectionSchema().dumps(result)
            return create_commercetools_response(request, text=content)
        return create_commercetools_response(request, status_code=404)

    def get_by_key(self, request, key):
        params = utils.parse_request_params(_ProductProjectionQuerySchema, request)
        obj = self.model.get_by_key(key)
        if obj:
            expanded_obj = self._expand(request, obj)
            product = ProductSchema().load(expanded_obj)

            result = self._convert_product_projection(product, params["staged"])
            if not result:
                return create_commercetools_response(request, status_code=404)
            content = ProductProjectionSchema().dumps(result)
            return create_commercetools_response(request, text=content)
        return create_commercetools_response(request, status_code=404)

    def _convert_product_projection(
        self, product: types.Product, staged: bool = False
    ) -> typing.Optional[types.ProductProjection]:
        """Convert a Product object to a ProductProjection object"""
        if product.master_data is None:
            return None

        if staged:
            data: types.ProductData = product.master_data.staged
        else:
            data: types.ProductData = product.master_data.current

        if data is None:
            return None

        return types.ProductProjection(
            id=product.id,
            version=product.version,
            key=product.key,
            created_at=product.created_at,
            last_modified_at=product.last_modified_at,
            product_type=product.product_type,
            name=data.name,
            description=data.description,
            slug=data.slug,
            categories=data.categories,
            category_order_hints=data.category_order_hints,
            meta_title=data.meta_title,
            meta_description=data.meta_description,
            meta_keywords=data.meta_keywords,
            search_keywords=data.search_keywords,
            has_staged_changes=product.master_data.has_staged_changes,
            published=product.master_data.published,
            master_variant=data.master_variant,
            variants=data.variants,
            tax_category=product.tax_category,
            state=product.state,
            review_rating_statistics=product.review_rating_statistics,
        )



# DO NOT EDIT! This file is automatically generated
import re
import typing

import marshmallow
import marshmallow_enum

from commercetools import helpers, types
from commercetools._schemas._common import (
    BaseResourceSchema,
    LocalizedStringField,
    ReferenceSchema,
    ResourceIdentifierSchema,
)
from commercetools._schemas._type import FieldContainerField

__all__ = [
    "AttributeSchema",
    "CustomTokenizerSchema",
    "FacetResultRangeSchema",
    "FacetResultSchema",
    "FacetResultTermSchema",
    "FacetResultsSchema",
    "FilteredFacetResultSchema",
    "ProductAddAssetActionSchema",
    "ProductAddExternalImageActionSchema",
    "ProductAddPriceActionSchema",
    "ProductAddToCategoryActionSchema",
    "ProductAddVariantActionSchema",
    "ProductCatalogDataSchema",
    "ProductChangeAssetNameActionSchema",
    "ProductChangeAssetOrderActionSchema",
    "ProductChangeMasterVariantActionSchema",
    "ProductChangeNameActionSchema",
    "ProductChangePriceActionSchema",
    "ProductChangeSlugActionSchema",
    "ProductDataSchema",
    "ProductDraftSchema",
    "ProductLegacySetSkuActionSchema",
    "ProductMoveImageToPositionActionSchema",
    "ProductPagedQueryResponseSchema",
    "ProductProjectionPagedQueryResponseSchema",
    "ProductProjectionPagedSearchResponseSchema",
    "ProductProjectionSchema",
    "ProductPublishActionSchema",
    "ProductReferenceSchema",
    "ProductRemoveAssetActionSchema",
    "ProductRemoveFromCategoryActionSchema",
    "ProductRemoveImageActionSchema",
    "ProductRemovePriceActionSchema",
    "ProductRemoveVariantActionSchema",
    "ProductResourceIdentifierSchema",
    "ProductRevertStagedChangesActionSchema",
    "ProductRevertStagedVariantChangesActionSchema",
    "ProductSchema",
    "ProductSetAssetCustomFieldActionSchema",
    "ProductSetAssetCustomTypeActionSchema",
    "ProductSetAssetDescriptionActionSchema",
    "ProductSetAssetKeyActionSchema",
    "ProductSetAssetSourcesActionSchema",
    "ProductSetAssetTagsActionSchema",
    "ProductSetAttributeActionSchema",
    "ProductSetAttributeInAllVariantsActionSchema",
    "ProductSetCategoryOrderHintActionSchema",
    "ProductSetDescriptionActionSchema",
    "ProductSetDiscountedPriceActionSchema",
    "ProductSetImageLabelActionSchema",
    "ProductSetKeyActionSchema",
    "ProductSetMetaDescriptionActionSchema",
    "ProductSetMetaKeywordsActionSchema",
    "ProductSetMetaTitleActionSchema",
    "ProductSetPricesActionSchema",
    "ProductSetProductPriceCustomFieldActionSchema",
    "ProductSetProductPriceCustomTypeActionSchema",
    "ProductSetProductVariantKeyActionSchema",
    "ProductSetSearchKeywordsActionSchema",
    "ProductSetSkuActionSchema",
    "ProductSetTaxCategoryActionSchema",
    "ProductTransitionStateActionSchema",
    "ProductUnpublishActionSchema",
    "ProductUpdateActionSchema",
    "ProductUpdateSchema",
    "ProductVariantAvailabilitySchema",
    "ProductVariantChannelAvailabilityMapSchema",
    "ProductVariantChannelAvailabilitySchema",
    "ProductVariantDraftSchema",
    "ProductVariantSchema",
    "RangeFacetResultSchema",
    "SearchKeywordSchema",
    "SearchKeywordsSchema",
    "SuggestTokenizerSchema",
    "SuggestionResultSchema",
    "SuggestionSchema",
    "TermFacetResultSchema",
    "WhitespaceTokenizerSchema",
]


class CategoryOrderHintsField(marshmallow.fields.Dict):
    def _deserialize(self, value, attr, data, **kwargs):
        result = super()._deserialize(value, attr, data)
        return types.CategoryOrderHints(**result)


class AttributeSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.Attribute`."""

    name = marshmallow.fields.String(allow_none=True)
    value = marshmallow.fields.Raw(allow_none=True)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.Attribute(**data)


class FacetResultRangeSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.FacetResultRange`."""

    from_ = marshmallow.fields.Integer(allow_none=True, data_key="from")
    from_str = marshmallow.fields.String(allow_none=True, data_key="fromStr")
    to = marshmallow.fields.Integer(allow_none=True)
    to_str = marshmallow.fields.String(allow_none=True, data_key="toStr")
    count = marshmallow.fields.Integer(allow_none=True)
    product_count = marshmallow.fields.Integer(
        allow_none=True, missing=None, data_key="productCount"
    )
    total = marshmallow.fields.Integer(allow_none=True)
    min = marshmallow.fields.Integer(allow_none=True)
    max = marshmallow.fields.Integer(allow_none=True)
    mean = marshmallow.fields.Integer(allow_none=True)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.FacetResultRange(**data)


class FacetResultSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.FacetResult`."""

    type = marshmallow_enum.EnumField(types.FacetTypes, by_value=True)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["type"]
        return types.FacetResult(**data)


class FacetResultTermSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.FacetResultTerm`."""

    term = marshmallow.fields.Raw(allow_none=True)
    count = marshmallow.fields.Integer(allow_none=True)
    product_count = marshmallow.fields.Integer(
        allow_none=True, missing=None, data_key="productCount"
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.FacetResultTerm(**data)


class FacetResultsSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.FacetResults`."""

    _regex = helpers.RegexField(
        unknown=marshmallow.EXCLUDE,
        pattern=re.compile("^[a-z].*$"),
        type=helpers.Discriminator(
            discriminator_field=("type", "type"),
            discriminator_schemas={
                "filter": "commercetools._schemas._product.FilteredFacetResultSchema",
                "range": "commercetools._schemas._product.RangeFacetResultSchema",
                "terms": "commercetools._schemas._product.TermFacetResultSchema",
            },
            unknown=marshmallow.EXCLUDE,
            allow_none=True,
        ),
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        data = typing.cast(helpers.RegexField, self.fields["_regex"]).postprocess(data)
        return types.FacetResults(**data)

    @marshmallow.pre_load
    def pre_load(self, data, **kwargs):
        data = typing.cast(helpers.RegexField, self.fields["_regex"]).preprocess(data)
        return data

    @marshmallow.pre_dump
    def pre_dump(self, data, **kwargs):
        data = typing.cast(helpers.RegexField, self.fields["_regex"]).preprocess(data)
        return data

    @marshmallow.post_dump
    def post_dump(self, data, **kwargs):
        data = typing.cast(helpers.RegexField, self.fields["_regex"]).postprocess(data)
        return data


class ProductCatalogDataSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.ProductCatalogData`."""

    published = marshmallow.fields.Bool(allow_none=True)
    current = helpers.LazyNestedField(
        nested="commercetools._schemas._product.ProductDataSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
    )
    staged = helpers.LazyNestedField(
        nested="commercetools._schemas._product.ProductDataSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
    )
    has_staged_changes = marshmallow.fields.Bool(
        allow_none=True, data_key="hasStagedChanges"
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.ProductCatalogData(**data)


class ProductDataSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.ProductData`."""

    name = LocalizedStringField(allow_none=True)
    categories = helpers.LazyNestedField(
        nested="commercetools._schemas._category.CategoryReferenceSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
    )
    category_order_hints = CategoryOrderHintsField(
        allow_none=True, missing=None, data_key="categoryOrderHints"
    )
    description = LocalizedStringField(allow_none=True, missing=None)
    slug = LocalizedStringField(allow_none=True)
    meta_title = LocalizedStringField(
        allow_none=True, missing=None, data_key="metaTitle"
    )
    meta_description = LocalizedStringField(
        allow_none=True, missing=None, data_key="metaDescription"
    )
    meta_keywords = LocalizedStringField(
        allow_none=True, missing=None, data_key="metaKeywords"
    )
    master_variant = helpers.LazyNestedField(
        nested="commercetools._schemas._product.ProductVariantSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        data_key="masterVariant",
    )
    variants = helpers.LazyNestedField(
        nested="commercetools._schemas._product.ProductVariantSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
    )
    search_keywords = helpers.LazyNestedField(
        nested="commercetools._schemas._product.SearchKeywordsSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        data_key="searchKeywords",
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.ProductData(**data)


class ProductDraftSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.ProductDraft`."""

    product_type = helpers.LazyNestedField(
        nested="commercetools._schemas._product_type.ProductTypeResourceIdentifierSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        data_key="productType",
    )
    name = LocalizedStringField(allow_none=True)
    slug = LocalizedStringField(allow_none=True)
    key = marshmallow.fields.String(allow_none=True, missing=None)
    description = LocalizedStringField(allow_none=True, missing=None)
    categories = helpers.LazyNestedField(
        nested="commercetools._schemas._category.CategoryResourceIdentifierSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
        missing=None,
    )
    category_order_hints = CategoryOrderHintsField(
        allow_none=True, missing=None, data_key="categoryOrderHints"
    )
    meta_title = LocalizedStringField(
        allow_none=True, missing=None, data_key="metaTitle"
    )
    meta_description = LocalizedStringField(
        allow_none=True, missing=None, data_key="metaDescription"
    )
    meta_keywords = LocalizedStringField(
        allow_none=True, missing=None, data_key="metaKeywords"
    )
    master_variant = helpers.LazyNestedField(
        nested="commercetools._schemas._product.ProductVariantDraftSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
        data_key="masterVariant",
    )
    variants = helpers.LazyNestedField(
        nested="commercetools._schemas._product.ProductVariantDraftSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
        missing=None,
    )
    tax_category = helpers.LazyNestedField(
        nested="commercetools._schemas._tax_category.TaxCategoryResourceIdentifierSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
        data_key="taxCategory",
    )
    search_keywords = helpers.LazyNestedField(
        nested="commercetools._schemas._product.SearchKeywordsSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
        data_key="searchKeywords",
    )
    state = helpers.LazyNestedField(
        nested="commercetools._schemas._state.StateResourceIdentifierSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
    )
    publish = marshmallow.fields.Bool(allow_none=True, missing=None)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.ProductDraft(**data)


class ProductPagedQueryResponseSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.ProductPagedQueryResponse`."""

    limit = marshmallow.fields.Integer(allow_none=True)
    count = marshmallow.fields.Integer(allow_none=True)
    total = marshmallow.fields.Integer(allow_none=True, missing=None)
    offset = marshmallow.fields.Integer(allow_none=True)
    results = helpers.LazyNestedField(
        nested="commercetools._schemas._product.ProductSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.ProductPagedQueryResponse(**data)


class ProductProjectionPagedQueryResponseSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.ProductProjectionPagedQueryResponse`."""

    limit = marshmallow.fields.Integer(allow_none=True)
    count = marshmallow.fields.Integer(allow_none=True)
    total = marshmallow.fields.Integer(allow_none=True, missing=None)
    offset = marshmallow.fields.Integer(allow_none=True)
    results = helpers.LazyNestedField(
        nested="commercetools._schemas._product.ProductProjectionSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.ProductProjectionPagedQueryResponse(**data)


class ProductProjectionPagedSearchResponseSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.ProductProjectionPagedSearchResponse`."""

    count = marshmallow.fields.Integer(allow_none=True)
    total = marshmallow.fields.Integer(allow_none=True, missing=None)
    offset = marshmallow.fields.Integer(allow_none=True)
    results = helpers.LazyNestedField(
        nested="commercetools._schemas._product.ProductProjectionSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
    )
    facets = helpers.LazyNestedField(
        nested="commercetools._schemas._product.FacetResultsSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.ProductProjectionPagedSearchResponse(**data)


class ProductProjectionSchema(BaseResourceSchema):
    """Marshmallow schema for :class:`commercetools.types.ProductProjection`."""

    id = marshmallow.fields.String(allow_none=True)
    version = marshmallow.fields.Integer(allow_none=True)
    key = marshmallow.fields.String(allow_none=True, missing=None)
    product_type = helpers.LazyNestedField(
        nested="commercetools._schemas._product_type.ProductTypeReferenceSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        data_key="productType",
    )
    name = LocalizedStringField(allow_none=True)
    description = LocalizedStringField(allow_none=True, missing=None)
    slug = LocalizedStringField(allow_none=True)
    categories = helpers.LazyNestedField(
        nested="commercetools._schemas._category.CategoryReferenceSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
    )
    category_order_hints = CategoryOrderHintsField(
        allow_none=True, missing=None, data_key="categoryOrderHints"
    )
    meta_title = LocalizedStringField(
        allow_none=True, missing=None, data_key="metaTitle"
    )
    meta_description = LocalizedStringField(
        allow_none=True, missing=None, data_key="metaDescription"
    )
    meta_keywords = LocalizedStringField(
        allow_none=True, missing=None, data_key="metaKeywords"
    )
    search_keywords = helpers.LazyNestedField(
        nested="commercetools._schemas._product.SearchKeywordsSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
        data_key="searchKeywords",
    )
    has_staged_changes = marshmallow.fields.Bool(
        allow_none=True, missing=None, data_key="hasStagedChanges"
    )
    published = marshmallow.fields.Bool(allow_none=True, missing=None)
    master_variant = helpers.LazyNestedField(
        nested="commercetools._schemas._product.ProductVariantSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        data_key="masterVariant",
    )
    variants = helpers.LazyNestedField(
        nested="commercetools._schemas._product.ProductVariantSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
    )
    tax_category = helpers.LazyNestedField(
        nested="commercetools._schemas._tax_category.TaxCategoryReferenceSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
        data_key="taxCategory",
    )
    state = helpers.LazyNestedField(
        nested="commercetools._schemas._state.StateReferenceSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
    )
    review_rating_statistics = helpers.LazyNestedField(
        nested="commercetools._schemas._review.ReviewRatingStatisticsSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
        data_key="reviewRatingStatistics",
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.ProductProjection(**data)


class ProductReferenceSchema(ReferenceSchema):
    """Marshmallow schema for :class:`commercetools.types.ProductReference`."""

    obj = helpers.LazyNestedField(
        nested="commercetools._schemas._product.ProductSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["type_id"]
        return types.ProductReference(**data)


class ProductResourceIdentifierSchema(ResourceIdentifierSchema):
    """Marshmallow schema for :class:`commercetools.types.ProductResourceIdentifier`."""

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["type_id"]
        return types.ProductResourceIdentifier(**data)


class ProductSchema(BaseResourceSchema):
    """Marshmallow schema for :class:`commercetools.types.Product`."""

    id = marshmallow.fields.String(allow_none=True)
    version = marshmallow.fields.Integer(allow_none=True)
    created_at = marshmallow.fields.DateTime(allow_none=True, data_key="createdAt")
    last_modified_at = marshmallow.fields.DateTime(
        allow_none=True, data_key="lastModifiedAt"
    )
    last_modified_by = helpers.LazyNestedField(
        nested="commercetools._schemas._common.LastModifiedBySchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
        data_key="lastModifiedBy",
    )
    created_by = helpers.LazyNestedField(
        nested="commercetools._schemas._common.CreatedBySchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
        data_key="createdBy",
    )
    key = marshmallow.fields.String(allow_none=True, missing=None)
    product_type = helpers.LazyNestedField(
        nested="commercetools._schemas._product_type.ProductTypeReferenceSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        data_key="productType",
    )
    master_data = helpers.LazyNestedField(
        nested="commercetools._schemas._product.ProductCatalogDataSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        data_key="masterData",
    )
    tax_category = helpers.LazyNestedField(
        nested="commercetools._schemas._tax_category.TaxCategoryReferenceSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
        data_key="taxCategory",
    )
    state = helpers.LazyNestedField(
        nested="commercetools._schemas._state.StateReferenceSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
    )
    review_rating_statistics = helpers.LazyNestedField(
        nested="commercetools._schemas._review.ReviewRatingStatisticsSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
        data_key="reviewRatingStatistics",
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.Product(**data)


class ProductUpdateActionSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.ProductUpdateAction`."""

    action = marshmallow.fields.String(allow_none=True)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.ProductUpdateAction(**data)


class ProductUpdateSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.ProductUpdate`."""

    version = marshmallow.fields.Integer(allow_none=True)
    actions = marshmallow.fields.List(
        helpers.Discriminator(
            discriminator_field=("action", "action"),
            discriminator_schemas={
                "addAsset": "commercetools._schemas._product.ProductAddAssetActionSchema",
                "addExternalImage": "commercetools._schemas._product.ProductAddExternalImageActionSchema",
                "addPrice": "commercetools._schemas._product.ProductAddPriceActionSchema",
                "addToCategory": "commercetools._schemas._product.ProductAddToCategoryActionSchema",
                "addVariant": "commercetools._schemas._product.ProductAddVariantActionSchema",
                "changeAssetName": "commercetools._schemas._product.ProductChangeAssetNameActionSchema",
                "changeAssetOrder": "commercetools._schemas._product.ProductChangeAssetOrderActionSchema",
                "changeMasterVariant": "commercetools._schemas._product.ProductChangeMasterVariantActionSchema",
                "changeName": "commercetools._schemas._product.ProductChangeNameActionSchema",
                "changePrice": "commercetools._schemas._product.ProductChangePriceActionSchema",
                "changeSlug": "commercetools._schemas._product.ProductChangeSlugActionSchema",
                "legacySetSku": "commercetools._schemas._product.ProductLegacySetSkuActionSchema",
                "moveImageToPosition": "commercetools._schemas._product.ProductMoveImageToPositionActionSchema",
                "publish": "commercetools._schemas._product.ProductPublishActionSchema",
                "removeAsset": "commercetools._schemas._product.ProductRemoveAssetActionSchema",
                "removeFromCategory": "commercetools._schemas._product.ProductRemoveFromCategoryActionSchema",
                "removeImage": "commercetools._schemas._product.ProductRemoveImageActionSchema",
                "removePrice": "commercetools._schemas._product.ProductRemovePriceActionSchema",
                "removeVariant": "commercetools._schemas._product.ProductRemoveVariantActionSchema",
                "revertStagedChanges": "commercetools._schemas._product.ProductRevertStagedChangesActionSchema",
                "revertStagedVariantChanges": "commercetools._schemas._product.ProductRevertStagedVariantChangesActionSchema",
                "setAssetCustomField": "commercetools._schemas._product.ProductSetAssetCustomFieldActionSchema",
                "setAssetCustomType": "commercetools._schemas._product.ProductSetAssetCustomTypeActionSchema",
                "setAssetDescription": "commercetools._schemas._product.ProductSetAssetDescriptionActionSchema",
                "setAssetKey": "commercetools._schemas._product.ProductSetAssetKeyActionSchema",
                "setAssetSources": "commercetools._schemas._product.ProductSetAssetSourcesActionSchema",
                "setAssetTags": "commercetools._schemas._product.ProductSetAssetTagsActionSchema",
                "setAttribute": "commercetools._schemas._product.ProductSetAttributeActionSchema",
                "setAttributeInAllVariants": "commercetools._schemas._product.ProductSetAttributeInAllVariantsActionSchema",
                "setCategoryOrderHint": "commercetools._schemas._product.ProductSetCategoryOrderHintActionSchema",
                "setDescription": "commercetools._schemas._product.ProductSetDescriptionActionSchema",
                "setDiscountedPrice": "commercetools._schemas._product.ProductSetDiscountedPriceActionSchema",
                "setImageLabel": "commercetools._schemas._product.ProductSetImageLabelActionSchema",
                "setKey": "commercetools._schemas._product.ProductSetKeyActionSchema",
                "setMetaDescription": "commercetools._schemas._product.ProductSetMetaDescriptionActionSchema",
                "setMetaKeywords": "commercetools._schemas._product.ProductSetMetaKeywordsActionSchema",
                "setMetaTitle": "commercetools._schemas._product.ProductSetMetaTitleActionSchema",
                "setPrices": "commercetools._schemas._product.ProductSetPricesActionSchema",
                "setProductPriceCustomField": "commercetools._schemas._product.ProductSetProductPriceCustomFieldActionSchema",
                "setProductPriceCustomType": "commercetools._schemas._product.ProductSetProductPriceCustomTypeActionSchema",
                "setProductVariantKey": "commercetools._schemas._product.ProductSetProductVariantKeyActionSchema",
                "setSearchKeywords": "commercetools._schemas._product.ProductSetSearchKeywordsActionSchema",
                "setSku": "commercetools._schemas._product.ProductSetSkuActionSchema",
                "setTaxCategory": "commercetools._schemas._product.ProductSetTaxCategoryActionSchema",
                "transitionState": "commercetools._schemas._product.ProductTransitionStateActionSchema",
                "unpublish": "commercetools._schemas._product.ProductUnpublishActionSchema",
            },
            unknown=marshmallow.EXCLUDE,
            allow_none=True,
        ),
        allow_none=True,
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.ProductUpdate(**data)


class ProductVariantAvailabilitySchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.ProductVariantAvailability`."""

    is_on_stock = marshmallow.fields.Bool(
        allow_none=True, missing=None, data_key="isOnStock"
    )
    restockable_in_days = marshmallow.fields.Integer(
        allow_none=True, missing=None, data_key="restockableInDays"
    )
    available_quantity = marshmallow.fields.Integer(
        allow_none=True, missing=None, data_key="availableQuantity"
    )
    channels = helpers.LazyNestedField(
        nested="commercetools._schemas._product.ProductVariantChannelAvailabilityMapSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.ProductVariantAvailability(**data)


class ProductVariantChannelAvailabilityMapSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.ProductVariantChannelAvailabilityMap`."""

    _regex = helpers.RegexField(
        unknown=marshmallow.EXCLUDE,
        pattern=re.compile(""),
        type=helpers.LazyNestedField(
            nested="commercetools._schemas._product.ProductVariantChannelAvailabilitySchema",
            unknown=marshmallow.EXCLUDE,
            allow_none=True,
        ),
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        data = typing.cast(helpers.RegexField, self.fields["_regex"]).postprocess(data)
        return types.ProductVariantChannelAvailabilityMap(**data)

    @marshmallow.pre_load
    def pre_load(self, data, **kwargs):
        data = typing.cast(helpers.RegexField, self.fields["_regex"]).preprocess(data)
        return data

    @marshmallow.pre_dump
    def pre_dump(self, data, **kwargs):
        data = typing.cast(helpers.RegexField, self.fields["_regex"]).preprocess(data)
        return data

    @marshmallow.post_dump
    def post_dump(self, data, **kwargs):
        data = typing.cast(helpers.RegexField, self.fields["_regex"]).postprocess(data)
        return data


class ProductVariantChannelAvailabilitySchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.ProductVariantChannelAvailability`."""

    is_on_stock = marshmallow.fields.Bool(
        allow_none=True, missing=None, data_key="isOnStock"
    )
    restockable_in_days = marshmallow.fields.Integer(
        allow_none=True, missing=None, data_key="restockableInDays"
    )
    available_quantity = marshmallow.fields.Integer(
        allow_none=True, missing=None, data_key="availableQuantity"
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.ProductVariantChannelAvailability(**data)


class ProductVariantDraftSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.ProductVariantDraft`."""

    sku = marshmallow.fields.String(allow_none=True, missing=None)
    key = marshmallow.fields.String(allow_none=True, missing=None)
    prices = helpers.LazyNestedField(
        nested="commercetools._schemas._common.PriceDraftSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
        missing=None,
    )
    attributes = helpers.LazyNestedField(
        nested="commercetools._schemas._product.AttributeSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
        missing=None,
    )
    images = helpers.LazyNestedField(
        nested="commercetools._schemas._common.ImageSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
        missing=None,
    )
    assets = helpers.LazyNestedField(
        nested="commercetools._schemas._common.AssetDraftSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
        missing=None,
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.ProductVariantDraft(**data)


class ProductVariantSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.ProductVariant`."""

    id = marshmallow.fields.Integer(allow_none=True)
    sku = marshmallow.fields.String(allow_none=True, missing=None)
    key = marshmallow.fields.String(allow_none=True, missing=None)
    prices = helpers.LazyNestedField(
        nested="commercetools._schemas._common.PriceSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
        missing=None,
    )
    attributes = helpers.LazyNestedField(
        nested="commercetools._schemas._product.AttributeSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
        missing=None,
    )
    price = helpers.LazyNestedField(
        nested="commercetools._schemas._common.PriceSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
    )
    images = helpers.LazyNestedField(
        nested="commercetools._schemas._common.ImageSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
        missing=None,
    )
    assets = helpers.LazyNestedField(
        nested="commercetools._schemas._common.AssetSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
        missing=None,
    )
    availability = helpers.LazyNestedField(
        nested="commercetools._schemas._product.ProductVariantAvailabilitySchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
    )
    is_matching_variant = marshmallow.fields.Bool(
        allow_none=True, missing=None, data_key="isMatchingVariant"
    )
    scoped_price = helpers.LazyNestedField(
        nested="commercetools._schemas._common.ScopedPriceSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
        data_key="scopedPrice",
    )
    scoped_price_discounted = marshmallow.fields.Bool(
        allow_none=True, missing=None, data_key="scopedPriceDiscounted"
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.ProductVariant(**data)


class SearchKeywordSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.SearchKeyword`."""

    text = marshmallow.fields.String(allow_none=True)
    suggest_tokenizer = helpers.Discriminator(
        discriminator_field=("type", "type"),
        discriminator_schemas={
            "custom": "commercetools._schemas._product.CustomTokenizerSchema",
            "whitespace": "commercetools._schemas._product.WhitespaceTokenizerSchema",
        },
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
        data_key="suggestTokenizer",
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.SearchKeyword(**data)


class SearchKeywordsSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.SearchKeywords`."""

    _regex = helpers.RegexField(
        unknown=marshmallow.EXCLUDE,
        pattern=re.compile("^[a-z]{2}(-[A-Z]{2})?$"),
        type=helpers.LazyNestedField(
            nested="commercetools._schemas._product.SearchKeywordSchema",
            unknown=marshmallow.EXCLUDE,
            allow_none=True,
            many=True,
        ),
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        data = typing.cast(helpers.RegexField, self.fields["_regex"]).postprocess(data)
        return types.SearchKeywords(**data)

    @marshmallow.pre_load
    def pre_load(self, data, **kwargs):
        data = typing.cast(helpers.RegexField, self.fields["_regex"]).preprocess(data)
        return data

    @marshmallow.pre_dump
    def pre_dump(self, data, **kwargs):
        data = typing.cast(helpers.RegexField, self.fields["_regex"]).preprocess(data)
        return data

    @marshmallow.post_dump
    def post_dump(self, data, **kwargs):
        data = typing.cast(helpers.RegexField, self.fields["_regex"]).postprocess(data)
        return data


class SuggestTokenizerSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.SuggestTokenizer`."""

    type = marshmallow.fields.String(allow_none=True)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["type"]
        return types.SuggestTokenizer(**data)


class SuggestionResultSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.SuggestionResult`."""

    _regex = helpers.RegexField(
        unknown=marshmallow.EXCLUDE,
        pattern=re.compile("searchKeywords.[a-z]{2}(-[A-Z]{2})?"),
        type=marshmallow.fields.List(
            helpers.LazyNestedField(
                nested="commercetools._schemas._product.SuggestionSchema",
                unknown=marshmallow.EXCLUDE,
                allow_none=True,
            ),
            allow_none=True,
        ),
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        data = typing.cast(helpers.RegexField, self.fields["_regex"]).postprocess(data)
        return types.SuggestionResult(**data)

    @marshmallow.pre_load
    def pre_load(self, data, **kwargs):
        data = typing.cast(helpers.RegexField, self.fields["_regex"]).preprocess(data)
        return data

    @marshmallow.pre_dump
    def pre_dump(self, data, **kwargs):
        data = typing.cast(helpers.RegexField, self.fields["_regex"]).preprocess(data)
        return data

    @marshmallow.post_dump
    def post_dump(self, data, **kwargs):
        data = typing.cast(helpers.RegexField, self.fields["_regex"]).postprocess(data)
        return data


class SuggestionSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.Suggestion`."""

    text = marshmallow.fields.String(allow_none=True)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.Suggestion(**data)


class CustomTokenizerSchema(SuggestTokenizerSchema):
    """Marshmallow schema for :class:`commercetools.types.CustomTokenizer`."""

    inputs = marshmallow.fields.List(marshmallow.fields.String(allow_none=True))

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["type"]
        return types.CustomTokenizer(**data)


class FilteredFacetResultSchema(FacetResultSchema):
    """Marshmallow schema for :class:`commercetools.types.FilteredFacetResult`."""

    count = marshmallow.fields.Integer(allow_none=True)
    product_count = marshmallow.fields.Integer(
        allow_none=True, missing=None, data_key="productCount"
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["type"]
        return types.FilteredFacetResult(**data)


class ProductAddAssetActionSchema(ProductUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.ProductAddAssetAction`."""

    variant_id = marshmallow.fields.Integer(
        allow_none=True, missing=None, data_key="variantId"
    )
    sku = marshmallow.fields.String(allow_none=True, missing=None)
    staged = marshmallow.fields.Bool(allow_none=True, missing=None)
    asset = helpers.LazyNestedField(
        nested="commercetools._schemas._common.AssetDraftSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
    )
    position = marshmallow.fields.Integer(allow_none=True, missing=None)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.ProductAddAssetAction(**data)


class ProductAddExternalImageActionSchema(ProductUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.ProductAddExternalImageAction`."""

    variant_id = marshmallow.fields.Integer(
        allow_none=True, missing=None, data_key="variantId"
    )
    sku = marshmallow.fields.String(allow_none=True, missing=None)
    image = helpers.LazyNestedField(
        nested="commercetools._schemas._common.ImageSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
    )
    staged = marshmallow.fields.Bool(allow_none=True, missing=None)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.ProductAddExternalImageAction(**data)


class ProductAddPriceActionSchema(ProductUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.ProductAddPriceAction`."""

    variant_id = marshmallow.fields.Integer(
        allow_none=True, missing=None, data_key="variantId"
    )
    sku = marshmallow.fields.String(allow_none=True, missing=None)
    price = helpers.LazyNestedField(
        nested="commercetools._schemas._common.PriceDraftSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
    )
    staged = marshmallow.fields.Bool(allow_none=True, missing=None)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.ProductAddPriceAction(**data)


class ProductAddToCategoryActionSchema(ProductUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.ProductAddToCategoryAction`."""

    category = helpers.LazyNestedField(
        nested="commercetools._schemas._category.CategoryResourceIdentifierSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
    )
    order_hint = marshmallow.fields.String(
        allow_none=True, missing=None, data_key="orderHint"
    )
    staged = marshmallow.fields.Bool(allow_none=True, missing=None)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.ProductAddToCategoryAction(**data)


class ProductAddVariantActionSchema(ProductUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.ProductAddVariantAction`."""

    sku = marshmallow.fields.String(allow_none=True, missing=None)
    key = marshmallow.fields.String(allow_none=True, missing=None)
    prices = helpers.LazyNestedField(
        nested="commercetools._schemas._common.PriceDraftSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
        missing=None,
    )
    images = helpers.LazyNestedField(
        nested="commercetools._schemas._common.ImageSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
        missing=None,
    )
    attributes = helpers.LazyNestedField(
        nested="commercetools._schemas._product.AttributeSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
        missing=None,
    )
    staged = marshmallow.fields.Bool(allow_none=True, missing=None)
    assets = helpers.LazyNestedField(
        nested="commercetools._schemas._common.AssetSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
        missing=None,
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.ProductAddVariantAction(**data)


class ProductChangeAssetNameActionSchema(ProductUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.ProductChangeAssetNameAction`."""

    variant_id = marshmallow.fields.Integer(
        allow_none=True, missing=None, data_key="variantId"
    )
    sku = marshmallow.fields.String(allow_none=True, missing=None)
    staged = marshmallow.fields.Bool(allow_none=True, missing=None)
    asset_id = marshmallow.fields.String(
        allow_none=True, missing=None, data_key="assetId"
    )
    asset_key = marshmallow.fields.String(
        allow_none=True, missing=None, data_key="assetKey"
    )
    name = LocalizedStringField(allow_none=True)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.ProductChangeAssetNameAction(**data)


class ProductChangeAssetOrderActionSchema(ProductUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.ProductChangeAssetOrderAction`."""

    variant_id = marshmallow.fields.Integer(
        allow_none=True, missing=None, data_key="variantId"
    )
    sku = marshmallow.fields.String(allow_none=True, missing=None)
    staged = marshmallow.fields.Bool(allow_none=True, missing=None)
    asset_order = marshmallow.fields.List(
        marshmallow.fields.String(allow_none=True), data_key="assetOrder"
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.ProductChangeAssetOrderAction(**data)


class ProductChangeMasterVariantActionSchema(ProductUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.ProductChangeMasterVariantAction`."""

    variant_id = marshmallow.fields.Integer(
        allow_none=True, missing=None, data_key="variantId"
    )
    sku = marshmallow.fields.String(allow_none=True, missing=None)
    staged = marshmallow.fields.Bool(allow_none=True, missing=None)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.ProductChangeMasterVariantAction(**data)


class ProductChangeNameActionSchema(ProductUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.ProductChangeNameAction`."""

    name = LocalizedStringField(allow_none=True)
    staged = marshmallow.fields.Bool(allow_none=True, missing=None)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.ProductChangeNameAction(**data)


class ProductChangePriceActionSchema(ProductUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.ProductChangePriceAction`."""

    price_id = marshmallow.fields.String(allow_none=True, data_key="priceId")
    price = helpers.LazyNestedField(
        nested="commercetools._schemas._common.PriceDraftSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
    )
    staged = marshmallow.fields.Bool(allow_none=True, missing=None)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.ProductChangePriceAction(**data)


class ProductChangeSlugActionSchema(ProductUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.ProductChangeSlugAction`."""

    slug = LocalizedStringField(allow_none=True)
    staged = marshmallow.fields.Bool(allow_none=True, missing=None)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.ProductChangeSlugAction(**data)


class ProductLegacySetSkuActionSchema(ProductUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.ProductLegacySetSkuAction`."""

    sku = marshmallow.fields.String(allow_none=True, missing=None)
    variant_id = marshmallow.fields.Integer(allow_none=True, data_key="variantId")

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.ProductLegacySetSkuAction(**data)


class ProductMoveImageToPositionActionSchema(ProductUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.ProductMoveImageToPositionAction`."""

    variant_id = marshmallow.fields.Integer(
        allow_none=True, missing=None, data_key="variantId"
    )
    sku = marshmallow.fields.String(allow_none=True, missing=None)
    image_url = marshmallow.fields.String(allow_none=True, data_key="imageUrl")
    position = marshmallow.fields.Integer(allow_none=True)
    staged = marshmallow.fields.Bool(allow_none=True, missing=None)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.ProductMoveImageToPositionAction(**data)


class ProductPublishActionSchema(ProductUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.ProductPublishAction`."""

    scope = marshmallow_enum.EnumField(
        types.ProductPublishScope, by_value=True, missing=None
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.ProductPublishAction(**data)


class ProductRemoveAssetActionSchema(ProductUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.ProductRemoveAssetAction`."""

    variant_id = marshmallow.fields.Integer(
        allow_none=True, missing=None, data_key="variantId"
    )
    sku = marshmallow.fields.String(allow_none=True, missing=None)
    staged = marshmallow.fields.Bool(allow_none=True, missing=None)
    asset_id = marshmallow.fields.String(
        allow_none=True, missing=None, data_key="assetId"
    )
    asset_key = marshmallow.fields.String(
        allow_none=True, missing=None, data_key="assetKey"
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.ProductRemoveAssetAction(**data)


class ProductRemoveFromCategoryActionSchema(ProductUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.ProductRemoveFromCategoryAction`."""

    category = helpers.LazyNestedField(
        nested="commercetools._schemas._category.CategoryResourceIdentifierSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
    )
    staged = marshmallow.fields.Bool(allow_none=True, missing=None)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.ProductRemoveFromCategoryAction(**data)


class ProductRemoveImageActionSchema(ProductUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.ProductRemoveImageAction`."""

    variant_id = marshmallow.fields.Integer(
        allow_none=True, missing=None, data_key="variantId"
    )
    sku = marshmallow.fields.String(allow_none=True, missing=None)
    image_url = marshmallow.fields.String(allow_none=True, data_key="imageUrl")
    staged = marshmallow.fields.Bool(allow_none=True, missing=None)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.ProductRemoveImageAction(**data)


class ProductRemovePriceActionSchema(ProductUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.ProductRemovePriceAction`."""

    price_id = marshmallow.fields.String(allow_none=True, data_key="priceId")
    staged = marshmallow.fields.Bool(allow_none=True, missing=None)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.ProductRemovePriceAction(**data)


class ProductRemoveVariantActionSchema(ProductUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.ProductRemoveVariantAction`."""

    id = marshmallow.fields.Integer(allow_none=True, missing=None)
    sku = marshmallow.fields.String(allow_none=True, missing=None)
    staged = marshmallow.fields.Bool(allow_none=True, missing=None)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.ProductRemoveVariantAction(**data)


class ProductRevertStagedChangesActionSchema(ProductUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.ProductRevertStagedChangesAction`."""

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.ProductRevertStagedChangesAction()


class ProductRevertStagedVariantChangesActionSchema(ProductUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.ProductRevertStagedVariantChangesAction`."""

    variant_id = marshmallow.fields.Integer(allow_none=True, data_key="variantId")

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.ProductRevertStagedVariantChangesAction(**data)


class ProductSetAssetCustomFieldActionSchema(ProductUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.ProductSetAssetCustomFieldAction`."""

    variant_id = marshmallow.fields.Integer(
        allow_none=True, missing=None, data_key="variantId"
    )
    sku = marshmallow.fields.String(allow_none=True, missing=None)
    staged = marshmallow.fields.Bool(allow_none=True, missing=None)
    asset_id = marshmallow.fields.String(
        allow_none=True, missing=None, data_key="assetId"
    )
    asset_key = marshmallow.fields.String(
        allow_none=True, missing=None, data_key="assetKey"
    )
    name = marshmallow.fields.String(allow_none=True)
    value = marshmallow.fields.Raw(allow_none=True, missing=None)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.ProductSetAssetCustomFieldAction(**data)


class ProductSetAssetCustomTypeActionSchema(ProductUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.ProductSetAssetCustomTypeAction`."""

    variant_id = marshmallow.fields.Integer(
        allow_none=True, missing=None, data_key="variantId"
    )
    sku = marshmallow.fields.String(allow_none=True, missing=None)
    staged = marshmallow.fields.Bool(allow_none=True, missing=None)
    asset_id = marshmallow.fields.String(
        allow_none=True, missing=None, data_key="assetId"
    )
    asset_key = marshmallow.fields.String(
        allow_none=True, missing=None, data_key="assetKey"
    )
    type = helpers.LazyNestedField(
        nested="commercetools._schemas._type.TypeResourceIdentifierSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
    )
    fields = marshmallow.fields.Dict(allow_none=True, missing=None)  # type: ignore

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.ProductSetAssetCustomTypeAction(**data)


class ProductSetAssetDescriptionActionSchema(ProductUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.ProductSetAssetDescriptionAction`."""

    variant_id = marshmallow.fields.Integer(
        allow_none=True, missing=None, data_key="variantId"
    )
    sku = marshmallow.fields.String(allow_none=True, missing=None)
    staged = marshmallow.fields.Bool(allow_none=True, missing=None)
    asset_id = marshmallow.fields.String(
        allow_none=True, missing=None, data_key="assetId"
    )
    asset_key = marshmallow.fields.String(
        allow_none=True, missing=None, data_key="assetKey"
    )
    description = LocalizedStringField(allow_none=True, missing=None)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.ProductSetAssetDescriptionAction(**data)


class ProductSetAssetKeyActionSchema(ProductUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.ProductSetAssetKeyAction`."""

    variant_id = marshmallow.fields.Integer(
        allow_none=True, missing=None, data_key="variantId"
    )
    sku = marshmallow.fields.String(allow_none=True, missing=None)
    staged = marshmallow.fields.Bool(allow_none=True, missing=None)
    asset_id = marshmallow.fields.String(allow_none=True, data_key="assetId")
    asset_key = marshmallow.fields.String(
        allow_none=True, missing=None, data_key="assetKey"
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.ProductSetAssetKeyAction(**data)


class ProductSetAssetSourcesActionSchema(ProductUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.ProductSetAssetSourcesAction`."""

    variant_id = marshmallow.fields.Integer(
        allow_none=True, missing=None, data_key="variantId"
    )
    sku = marshmallow.fields.String(allow_none=True, missing=None)
    staged = marshmallow.fields.Bool(allow_none=True, missing=None)
    asset_id = marshmallow.fields.String(
        allow_none=True, missing=None, data_key="assetId"
    )
    asset_key = marshmallow.fields.String(
        allow_none=True, missing=None, data_key="assetKey"
    )
    sources = helpers.LazyNestedField(
        nested="commercetools._schemas._common.AssetSourceSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.ProductSetAssetSourcesAction(**data)


class ProductSetAssetTagsActionSchema(ProductUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.ProductSetAssetTagsAction`."""

    variant_id = marshmallow.fields.Integer(
        allow_none=True, missing=None, data_key="variantId"
    )
    sku = marshmallow.fields.String(allow_none=True, missing=None)
    staged = marshmallow.fields.Bool(allow_none=True, missing=None)
    asset_id = marshmallow.fields.String(
        allow_none=True, missing=None, data_key="assetId"
    )
    asset_key = marshmallow.fields.String(
        allow_none=True, missing=None, data_key="assetKey"
    )
    tags = marshmallow.fields.List(
        marshmallow.fields.String(allow_none=True), missing=None
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.ProductSetAssetTagsAction(**data)


class ProductSetAttributeActionSchema(ProductUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.ProductSetAttributeAction`."""

    variant_id = marshmallow.fields.Integer(
        allow_none=True, missing=None, data_key="variantId"
    )
    sku = marshmallow.fields.String(allow_none=True, missing=None)
    name = marshmallow.fields.String(allow_none=True)
    value = marshmallow.fields.Raw(allow_none=True, missing=None)
    staged = marshmallow.fields.Bool(allow_none=True, missing=None)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.ProductSetAttributeAction(**data)


class ProductSetAttributeInAllVariantsActionSchema(ProductUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.ProductSetAttributeInAllVariantsAction`."""

    name = marshmallow.fields.String(allow_none=True)
    value = marshmallow.fields.Raw(allow_none=True, missing=None)
    staged = marshmallow.fields.Bool(allow_none=True, missing=None)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.ProductSetAttributeInAllVariantsAction(**data)


class ProductSetCategoryOrderHintActionSchema(ProductUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.ProductSetCategoryOrderHintAction`."""

    category_id = marshmallow.fields.String(allow_none=True, data_key="categoryId")
    order_hint = marshmallow.fields.String(
        allow_none=True, missing=None, data_key="orderHint"
    )
    staged = marshmallow.fields.Bool(allow_none=True, missing=None)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.ProductSetCategoryOrderHintAction(**data)


class ProductSetDescriptionActionSchema(ProductUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.ProductSetDescriptionAction`."""

    description = LocalizedStringField(allow_none=True, missing=None)
    staged = marshmallow.fields.Bool(allow_none=True, missing=None)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.ProductSetDescriptionAction(**data)


class ProductSetDiscountedPriceActionSchema(ProductUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.ProductSetDiscountedPriceAction`."""

    price_id = marshmallow.fields.String(allow_none=True, data_key="priceId")
    staged = marshmallow.fields.Bool(allow_none=True, missing=None)
    discounted = helpers.LazyNestedField(
        nested="commercetools._schemas._common.DiscountedPriceSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.ProductSetDiscountedPriceAction(**data)


class ProductSetImageLabelActionSchema(ProductUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.ProductSetImageLabelAction`."""

    sku = marshmallow.fields.String(allow_none=True, missing=None)
    variant_id = marshmallow.fields.Integer(
        allow_none=True, missing=None, data_key="variantId"
    )
    image_url = marshmallow.fields.String(allow_none=True, data_key="imageUrl")
    label = marshmallow.fields.String(allow_none=True, missing=None)
    staged = marshmallow.fields.Bool(allow_none=True, missing=None)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.ProductSetImageLabelAction(**data)


class ProductSetKeyActionSchema(ProductUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.ProductSetKeyAction`."""

    key = marshmallow.fields.String(allow_none=True, missing=None)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.ProductSetKeyAction(**data)


class ProductSetMetaDescriptionActionSchema(ProductUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.ProductSetMetaDescriptionAction`."""

    meta_description = LocalizedStringField(
        allow_none=True, missing=None, data_key="metaDescription"
    )
    staged = marshmallow.fields.Bool(allow_none=True, missing=None)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.ProductSetMetaDescriptionAction(**data)


class ProductSetMetaKeywordsActionSchema(ProductUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.ProductSetMetaKeywordsAction`."""

    meta_keywords = LocalizedStringField(
        allow_none=True, missing=None, data_key="metaKeywords"
    )
    staged = marshmallow.fields.Bool(allow_none=True, missing=None)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.ProductSetMetaKeywordsAction(**data)


class ProductSetMetaTitleActionSchema(ProductUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.ProductSetMetaTitleAction`."""

    meta_title = LocalizedStringField(
        allow_none=True, missing=None, data_key="metaTitle"
    )
    staged = marshmallow.fields.Bool(allow_none=True, missing=None)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.ProductSetMetaTitleAction(**data)


class ProductSetPricesActionSchema(ProductUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.ProductSetPricesAction`."""

    variant_id = marshmallow.fields.Integer(
        allow_none=True, missing=None, data_key="variantId"
    )
    sku = marshmallow.fields.String(allow_none=True, missing=None)
    prices = helpers.LazyNestedField(
        nested="commercetools._schemas._common.PriceDraftSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
    )
    staged = marshmallow.fields.Bool(allow_none=True, missing=None)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.ProductSetPricesAction(**data)


class ProductSetProductPriceCustomFieldActionSchema(ProductUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.ProductSetProductPriceCustomFieldAction`."""

    price_id = marshmallow.fields.String(allow_none=True, data_key="priceId")
    staged = marshmallow.fields.Bool(allow_none=True, missing=None)
    name = marshmallow.fields.String(allow_none=True)
    value = marshmallow.fields.Raw(allow_none=True, missing=None)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.ProductSetProductPriceCustomFieldAction(**data)


class ProductSetProductPriceCustomTypeActionSchema(ProductUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.ProductSetProductPriceCustomTypeAction`."""

    price_id = marshmallow.fields.String(allow_none=True, data_key="priceId")
    staged = marshmallow.fields.Bool(allow_none=True, missing=None)
    type = helpers.LazyNestedField(
        nested="commercetools._schemas._type.TypeResourceIdentifierSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
    )
    fields = FieldContainerField(allow_none=True, missing=None)  # type: ignore

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.ProductSetProductPriceCustomTypeAction(**data)


class ProductSetProductVariantKeyActionSchema(ProductUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.ProductSetProductVariantKeyAction`."""

    variant_id = marshmallow.fields.Integer(
        allow_none=True, missing=None, data_key="variantId"
    )
    sku = marshmallow.fields.String(allow_none=True, missing=None)
    key = marshmallow.fields.String(allow_none=True, missing=None)
    staged = marshmallow.fields.Bool(allow_none=True, missing=None)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.ProductSetProductVariantKeyAction(**data)


class ProductSetSearchKeywordsActionSchema(ProductUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.ProductSetSearchKeywordsAction`."""

    search_keywords = helpers.LazyNestedField(
        nested="commercetools._schemas._product.SearchKeywordsSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        data_key="searchKeywords",
    )
    staged = marshmallow.fields.Bool(allow_none=True, missing=None)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.ProductSetSearchKeywordsAction(**data)


class ProductSetSkuActionSchema(ProductUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.ProductSetSkuAction`."""

    variant_id = marshmallow.fields.Integer(allow_none=True, data_key="variantId")
    sku = marshmallow.fields.String(allow_none=True, missing=None)
    staged = marshmallow.fields.Bool(allow_none=True, missing=None)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.ProductSetSkuAction(**data)


class ProductSetTaxCategoryActionSchema(ProductUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.ProductSetTaxCategoryAction`."""

    tax_category = helpers.LazyNestedField(
        nested="commercetools._schemas._tax_category.TaxCategoryResourceIdentifierSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
        data_key="taxCategory",
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.ProductSetTaxCategoryAction(**data)


class ProductTransitionStateActionSchema(ProductUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.ProductTransitionStateAction`."""

    state = helpers.LazyNestedField(
        nested="commercetools._schemas._state.StateResourceIdentifierSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
    )
    force = marshmallow.fields.Bool(allow_none=True, missing=None)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.ProductTransitionStateAction(**data)


class ProductUnpublishActionSchema(ProductUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.ProductUnpublishAction`."""

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.ProductUnpublishAction()


class RangeFacetResultSchema(FacetResultSchema):
    """Marshmallow schema for :class:`commercetools.types.RangeFacetResult`."""

    ranges = helpers.LazyNestedField(
        nested="commercetools._schemas._product.FacetResultRangeSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["type"]
        return types.RangeFacetResult(**data)


class TermFacetResultSchema(FacetResultSchema):
    """Marshmallow schema for :class:`commercetools.types.TermFacetResult`."""

    data_type = marshmallow_enum.EnumField(
        types.TermFacetResultType, by_value=True, data_key="dataType"
    )
    missing = marshmallow.fields.Integer(allow_none=True)
    total = marshmallow.fields.Integer(allow_none=True)
    other = marshmallow.fields.Integer(allow_none=True)
    terms = helpers.LazyNestedField(
        nested="commercetools._schemas._product.FacetResultTermSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["type"]
        return types.TermFacetResult(**data)


class WhitespaceTokenizerSchema(SuggestTokenizerSchema):
    """Marshmallow schema for :class:`commercetools.types.WhitespaceTokenizer`."""

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["type"]
        return types.WhitespaceTokenizer()



import copy
import datetime
import uuid
from typing import Dict, List, Optional

from commercetools import types
from commercetools._schemas._channel import (
    ChannelResourceIdentifierSchema,
    ChannelSchema,
)
from commercetools._schemas._store import (
    StoreDraftSchema,
    StorePagedQueryResponseSchema,
    StoreSchema,
    StoreUpdateSchema,
)
from commercetools.testing.abstract import BaseModel, ServiceBackend
from commercetools.testing.utils import InternalUpdateError, update_attribute


class StoresModel(BaseModel):
    _primary_type_name = "store"
    _resource_schema = StoreSchema
    _unique_values = ["key"]

    def _create_from_draft(
        self, draft: types.StoreDraft, id: Optional[str] = None
    ) -> types.Store:
        object_id = str(uuid.UUID(id) if id is not None else uuid.uuid4())
        distribution_channels: List[types.ChannelReference] = []
        if draft.distribution_channels:
            distribution_channels = convert_identifiers_to_references(
                self._storage._stores["channel"], draft.distribution_channels
            )

        return types.Store(
            id=str(object_id),
            created_at=datetime.datetime.now(datetime.timezone.utc),
            last_modified_at=datetime.datetime.now(datetime.timezone.utc),
            version=1,
            key=draft.key,
            name=draft.name,
            languages=draft.languages,
            distribution_channels=distribution_channels,
        )


def convert_identifiers_to_references(
    channel_store: Dict, channel_identifiers: List[types.ChannelResourceIdentifier]
) -> List[types.ChannelReference]:
    channel_references: List[types.ChannelReference] = []
    for ci in channel_identifiers:
        channel_data: Optional[Dict] = None
        for c in channel_store.values():
            if ci.key and c["key"] == ci.key:
                channel_data = c
                break
            if ci.id and c["id"] == ci.id:
                channel_data = c
                break
        if not channel_data:
            raise InternalUpdateError("Channel not found.")
        channel: types.Channel = ChannelSchema().load(data=channel_data)
        if types.ChannelRoleEnum.PRODUCT_DISTRIBUTION not in channel.roles:
            raise InternalUpdateError(
                "Channel does not have product distribution role."
            )
        channel_references.append(types.ChannelReference(id=channel.id))
    return channel_references


def set_languages():
    def updater(self, obj, action):
        value = getattr(action, "languages")
        new = copy.deepcopy(obj)
        new["languages"] = value
        return new

    return updater


def set_distribution_channels(
    backend: "StoresBackend",
    obj: Dict,
    action: types.StoresSetDistributionChannelsAction,
):
    channel_references = []
    if action.distribution_channels:
        channel_references = convert_identifiers_to_references(
            backend.model._storage._stores["channel"], action.distribution_channels
        )
    new = copy.deepcopy(obj)
    new["distributionChannels"] = ChannelResourceIdentifierSchema().dump(
        channel_references, many=True
    )
    return new


class StoresBackend(ServiceBackend):
    service_path = "stores"
    model_class = StoresModel
    _schema_draft = StoreDraftSchema
    _schema_query_response = StorePagedQueryResponseSchema
    _schema_update = StoreUpdateSchema
    _verify_version = False

    def urls(self):
        return [
            ("^$", "GET", self.query),
            ("^$", "POST", self.create),
            ("^(?P<id>[^/]+)$", "GET", self.get_by_id),
            ("^(?P<id>[^/]+)$", "DELETE", self.delete_by_id),
            ("^(?P<id>[^/]+)$", "POST", self.update_by_id),
            ("^key=(?P<key>[^/]+)$", "GET", self.get_by_key),
            ("^key=(?P<key>[^/]+)$", "DELETE", self.delete_by_key),
            ("^key=(?P<key>[^/]+)$", "POST", self.update_by_key),
        ]

    _actions = {
        "setName": update_attribute("name", "name"),
        "setLanguages": set_languages(),
        "setDistributionChannels": set_distribution_channels,
    }



# DO NOT EDIT! This file is automatically generated
import typing

from commercetools._schemas._subscription import (
    SubscriptionDraftSchema,
    SubscriptionPagedQueryResponseSchema,
    SubscriptionSchema,
    SubscriptionUpdateSchema,
)
from commercetools.helpers import RemoveEmptyValuesMixin
from commercetools.types._subscription import (
    Subscription,
    SubscriptionDraft,
    SubscriptionPagedQueryResponse,
    SubscriptionUpdate,
    SubscriptionUpdateAction,
)
from commercetools.typing import OptionalListStr

from . import abstract, traits


class _SubscriptionQuerySchema(
    traits.ExpandableSchema,
    traits.SortableSchema,
    traits.PagingSchema,
    traits.QuerySchema,
):
    pass


class _SubscriptionUpdateSchema(traits.ExpandableSchema, traits.VersionedSchema):
    pass


class _SubscriptionDeleteSchema(traits.VersionedSchema, traits.ExpandableSchema):
    pass


class SubscriptionService(abstract.AbstractService):
    """Subscriptions allow you to be notified of new messages or changes via a
    Message Queue of your choice"""

    def get_by_id(self, id: str, *, expand: OptionalListStr = None) -> Subscription:
        """Retrieves the representation of a subscription by its id."""
        params = self._serialize_params({"expand": expand}, traits.ExpandableSchema)
        return self._client._get(
            endpoint=f"subscriptions/{id}", params=params, schema_cls=SubscriptionSchema
        )

    def get_by_key(self, key: str, *, expand: OptionalListStr = None) -> Subscription:
        """Retrieves the representation of a subscription by its key."""
        params = self._serialize_params({"expand": expand}, traits.ExpandableSchema)
        return self._client._get(
            endpoint=f"subscriptions/key={key}",
            params=params,
            schema_cls=SubscriptionSchema,
        )

    def query(
        self,
        *,
        expand: OptionalListStr = None,
        sort: OptionalListStr = None,
        limit: int = None,
        offset: int = None,
        with_total: bool = None,
        where: OptionalListStr = None,
        predicate_var: typing.Dict[str, str] = None,
    ) -> SubscriptionPagedQueryResponse:
        """Subscriptions allow you to be notified of new messages or changes via a
        Message Queue of your choice
        """
        params = self._serialize_params(
            {
                "expand": expand,
                "sort": sort,
                "limit": limit,
                "offset": offset,
                "with_total": with_total,
                "where": where,
                "predicate_var": predicate_var,
            },
            _SubscriptionQuerySchema,
        )
        return self._client._get(
            endpoint="subscriptions",
            params=params,
            schema_cls=SubscriptionPagedQueryResponseSchema,
        )

    def create(
        self, draft: SubscriptionDraft, *, expand: OptionalListStr = None
    ) -> Subscription:
        """The creation of a Subscription is eventually consistent, it may take up
        to a minute before it becomes fully active.

        In order to test that the destination is correctly configured, a test
        message will be put into the queue. If the message could not be
        delivered, the subscription will not be created. The payload of the test
        message is a notification of type ResourceCreated for the resourceTypeId
        subscription. Currently, a maximum of 25 subscriptions can be created per
        project.   Subscriptions allow you to be notified of new messages or
        changes via a Message Queue of your choice
        """
        params = self._serialize_params({"expand": expand}, traits.ExpandableSchema)
        return self._client._post(
            endpoint="subscriptions",
            params=params,
            data_object=draft,
            request_schema_cls=SubscriptionDraftSchema,
            response_schema_cls=SubscriptionSchema,
        )

    def update_by_id(
        self,
        id: str,
        version: int,
        actions: typing.List[SubscriptionUpdateAction],
        *,
        expand: OptionalListStr = None,
        force_update: bool = False,
    ) -> Subscription:
        params = self._serialize_params({"expand": expand}, _SubscriptionUpdateSchema)
        update_action = SubscriptionUpdate(version=version, actions=actions)
        return self._client._post(
            endpoint=f"subscriptions/{id}",
            params=params,
            data_object=update_action,
            request_schema_cls=SubscriptionUpdateSchema,
            response_schema_cls=SubscriptionSchema,
            force_update=force_update,
        )

    def update_by_key(
        self,
        key: str,
        version: int,
        actions: typing.List[SubscriptionUpdateAction],
        *,
        expand: OptionalListStr = None,
        force_update: bool = False,
    ) -> Subscription:
        params = self._serialize_params({"expand": expand}, _SubscriptionUpdateSchema)
        update_action = SubscriptionUpdate(version=version, actions=actions)
        return self._client._post(
            endpoint=f"subscriptions/key={key}",
            params=params,
            data_object=update_action,
            request_schema_cls=SubscriptionUpdateSchema,
            response_schema_cls=SubscriptionSchema,
            force_update=force_update,
        )

    def delete_by_id(
        self,
        id: str,
        version: int,
        *,
        expand: OptionalListStr = None,
        force_delete: bool = False,
    ) -> Subscription:
        params = self._serialize_params(
            {"version": version, "expand": expand}, _SubscriptionDeleteSchema
        )
        return self._client._delete(
            endpoint=f"subscriptions/{id}",
            params=params,
            response_schema_cls=SubscriptionSchema,
            force_delete=force_delete,
        )

    def delete_by_key(
        self,
        key: str,
        version: int,
        *,
        expand: OptionalListStr = None,
        force_delete: bool = False,
    ) -> Subscription:
        params = self._serialize_params(
            {"version": version, "expand": expand}, _SubscriptionDeleteSchema
        )
        return self._client._delete(
            endpoint=f"subscriptions/key={key}",
            params=params,
            response_schema_cls=SubscriptionSchema,
            force_delete=force_delete,
        )



import pytest
from requests.exceptions import HTTPError

from commercetools import types


def test_discount_code_get_by_id(client):
    discount_code = client.discount_codes.create(
        types.DiscountCodeDraft(
            name=types.LocalizedString({"en": "test discount"}),
            code="1337",
            cart_discounts=[],
        )
    )

    assert discount_code.id
    assert discount_code.code == "1337"

    discount_code = client.discount_codes.get_by_id(discount_code.id)
    assert discount_code.id
    assert discount_code.code == "1337"

    with pytest.raises(HTTPError):
        client.discount_codes.get_by_id("invalid")


def test_discount_code_query(client):
    client.discount_codes.create(
        types.DiscountCodeDraft(
            name=types.LocalizedString({"en:": "test discount"}),
            code="1337",
            cart_discounts=[],
        )
    )
    client.discount_codes.create(
        types.DiscountCodeDraft(
            name=types.LocalizedString({"en:": "test discount"}),
            code="1338",
            cart_discounts=[],
        )
    )

    # single sort query
    result = client.discount_codes.query(sort="id asc")
    assert len(result.results) == 2
    assert result.total == 2

    # multiple sort queries
    result = client.discount_codes.query(sort=["id asc", "name asc"])
    assert len(result.results) == 2
    assert result.total == 2


def test_discount_code_update(client):
    discount_code = client.discount_codes.create(
        types.DiscountCodeDraft(
            name=types.LocalizedString(en="en-discount_code"),
            code="1337",
            is_active=True,
            cart_discounts=[],
        )
    )
    assert discount_code.code == "1337"

    discount_code = client.discount_codes.update_by_id(
        id=discount_code.id,
        version=discount_code.version,
        actions=[types.DiscountCodeChangeIsActiveAction(is_active=False)],
    )

    assert discount_code.is_active is False



# DO NOT EDIT! This file is automatically generated
import datetime
import typing

from commercetools.types._abstract import _BaseType

if typing.TYPE_CHECKING:
    from ._channel import ChannelReference
    from ._common import LocalizedString, Price, PriceDraft, Reference, ReferenceTypeId
    from ._customer_group import CustomerGroupReference
    from ._product import Attribute
__all__ = [
    "AccessDeniedError",
    "ConcurrentModificationError",
    "DiscountCodeNonApplicableError",
    "DuplicateAttributeValueError",
    "DuplicateAttributeValuesError",
    "DuplicateFieldError",
    "DuplicateFieldWithConflictingResourceError",
    "DuplicatePriceScopeError",
    "DuplicateVariantValuesError",
    "EnumValueIsUsedError",
    "ErrorByExtension",
    "ErrorObject",
    "ErrorResponse",
    "ExtensionBadResponseError",
    "ExtensionNoResponseError",
    "ExtensionUpdateActionsFailedError",
    "InsufficientScopeError",
    "InvalidCredentialsError",
    "InvalidCurrentPasswordError",
    "InvalidFieldError",
    "InvalidInputError",
    "InvalidItemShippingDetailsError",
    "InvalidJsonInputError",
    "InvalidOperationError",
    "InvalidSubjectError",
    "InvalidTokenError",
    "LanguageUsedInStoresError",
    "MatchingPriceNotFoundError",
    "MissingTaxRateForCountryError",
    "NoMatchingProductDiscountFoundError",
    "OutOfStockError",
    "PriceChangedError",
    "ReferenceExistsError",
    "ReferencedResourceNotFoundError",
    "RequiredFieldError",
    "ResourceNotFoundError",
    "ShippingMethodDoesNotMatchCartError",
    "VariantValues",
]


class ErrorByExtension(_BaseType):
    #: :class:`str`
    id: str
    #: Optional :class:`str`
    key: typing.Optional[str]

    def __init__(self, *, id: str, key: typing.Optional[str] = None) -> None:
        self.id = id
        self.key = key
        super().__init__()

    def __repr__(self) -> str:
        return "ErrorByExtension(id=%r, key=%r)" % (self.id, self.key)


class ErrorObject(_BaseType):
    #: :class:`str`
    code: str
    #: :class:`str`
    message: str

    def __init__(self, *, code: str, message: str) -> None:
        self.code = code
        self.message = message
        super().__init__()

    def __repr__(self) -> str:
        return "ErrorObject(code=%r, message=%r)" % (self.code, self.message)


class ErrorResponse(_BaseType):
    #: :class:`int` `(Named` ``statusCode`` `in Commercetools)`
    status_code: int
    #: :class:`str`
    message: str
    #: Optional :class:`str`
    error: typing.Optional[str]
    #: Optional :class:`str`
    error_description: typing.Optional[str]
    #: Optional :class:`list`
    errors: typing.Optional[list]

    def __init__(
        self,
        *,
        status_code: int,
        message: str,
        error: typing.Optional[str] = None,
        error_description: typing.Optional[str] = None,
        errors: typing.Optional[list] = None
    ) -> None:
        self.status_code = status_code
        self.message = message
        self.error = error
        self.error_description = error_description
        self.errors = errors
        super().__init__()

    def __repr__(self) -> str:
        return (
            "ErrorResponse(status_code=%r, message=%r, error=%r, error_description=%r, errors=%r)"
            % (
                self.status_code,
                self.message,
                self.error,
                self.error_description,
                self.errors,
            )
        )


class VariantValues(_BaseType):
    #: Optional :class:`str`
    sku: typing.Optional[str]
    #: List of :class:`commercetools.types.PriceDraft`
    prices: typing.List["PriceDraft"]
    #: List of :class:`commercetools.types.Attribute`
    attributes: typing.List["Attribute"]

    def __init__(
        self,
        *,
        prices: typing.List["PriceDraft"],
        attributes: typing.List["Attribute"],
        sku: typing.Optional[str] = None
    ) -> None:
        self.sku = sku
        self.prices = prices
        self.attributes = attributes
        super().__init__()

    def __repr__(self) -> str:
        return "VariantValues(sku=%r, prices=%r, attributes=%r)" % (
            self.sku,
            self.prices,
            self.attributes,
        )


class AccessDeniedError(ErrorObject):
    def __init__(self, *, message: str) -> None:
        super().__init__(code="access_denied", message=message)

    def __repr__(self) -> str:
        return "AccessDeniedError(code=%r, message=%r)" % (self.code, self.message)


class ConcurrentModificationError(ErrorObject):
    #: Optional :class:`int` `(Named` ``currentVersion`` `in Commercetools)`
    current_version: typing.Optional[int]

    def __init__(
        self, *, message: str, current_version: typing.Optional[int] = None
    ) -> None:
        self.current_version = current_version
        super().__init__(code="ConcurrentModification", message=message)

    def __repr__(self) -> str:
        return (
            "ConcurrentModificationError(code=%r, message=%r, current_version=%r)"
            % (self.code, self.message, self.current_version)
        )


class DiscountCodeNonApplicableError(ErrorObject):
    #: Optional :class:`str` `(Named` ``discountCode`` `in Commercetools)`
    discount_code: typing.Optional[str]
    #: Optional :class:`str`
    reason: typing.Optional[str]
    #: Optional :class:`str` `(Named` ``dicountCodeId`` `in Commercetools)`
    dicount_code_id: typing.Optional[str]
    #: Optional :class:`datetime.datetime` `(Named` ``validFrom`` `in Commercetools)`
    valid_from: typing.Optional[datetime.datetime]
    #: Optional :class:`datetime.datetime` `(Named` ``validUntil`` `in Commercetools)`
    valid_until: typing.Optional[datetime.datetime]
    #: Optional :class:`datetime.datetime` `(Named` ``validityCheckTime`` `in Commercetools)`
    validity_check_time: typing.Optional[datetime.datetime]

    def __init__(
        self,
        *,
        message: str,
        discount_code: typing.Optional[str] = None,
        reason: typing.Optional[str] = None,
        dicount_code_id: typing.Optional[str] = None,
        valid_from: typing.Optional[datetime.datetime] = None,
        valid_until: typing.Optional[datetime.datetime] = None,
        validity_check_time: typing.Optional[datetime.datetime] = None
    ) -> None:
        self.discount_code = discount_code
        self.reason = reason
        self.dicount_code_id = dicount_code_id
        self.valid_from = valid_from
        self.valid_until = valid_until
        self.validity_check_time = validity_check_time
        super().__init__(code="DiscountCodeNonApplicable", message=message)

    def __repr__(self) -> str:
        return (
            "DiscountCodeNonApplicableError(code=%r, message=%r, discount_code=%r, reason=%r, dicount_code_id=%r, valid_from=%r, valid_until=%r, validity_check_time=%r)"
            % (
                self.code,
                self.message,
                self.discount_code,
                self.reason,
                self.dicount_code_id,
                self.valid_from,
                self.valid_until,
                self.validity_check_time,
            )
        )


class DuplicateAttributeValueError(ErrorObject):
    #: :class:`commercetools.types.Attribute`
    attribute: "Attribute"

    def __init__(self, *, message: str, attribute: "Attribute") -> None:
        self.attribute = attribute
        super().__init__(code="DuplicateAttributeValue", message=message)

    def __repr__(self) -> str:
        return "DuplicateAttributeValueError(code=%r, message=%r, attribute=%r)" % (
            self.code,
            self.message,
            self.attribute,
        )


class DuplicateAttributeValuesError(ErrorObject):
    #: List of :class:`commercetools.types.Attribute`
    attributes: typing.List["Attribute"]

    def __init__(self, *, message: str, attributes: typing.List["Attribute"]) -> None:
        self.attributes = attributes
        super().__init__(code="DuplicateAttributeValues", message=message)

    def __repr__(self) -> str:
        return "DuplicateAttributeValuesError(code=%r, message=%r, attributes=%r)" % (
            self.code,
            self.message,
            self.attributes,
        )


class DuplicateFieldError(ErrorObject):
    #: Optional :class:`str`
    field: typing.Optional[str]
    #: Optional :class:`typing.Any` `(Named` ``duplicateValue`` `in Commercetools)`
    duplicate_value: typing.Optional[typing.Any]
    #: Optional :class:`commercetools.types.Reference` `(Named` ``conflictingResource`` `in Commercetools)`
    conflicting_resource: typing.Optional["Reference"]

    def __init__(
        self,
        *,
        message: str,
        field: typing.Optional[str] = None,
        duplicate_value: typing.Optional[typing.Any] = None,
        conflicting_resource: typing.Optional["Reference"] = None
    ) -> None:
        self.field = field
        self.duplicate_value = duplicate_value
        self.conflicting_resource = conflicting_resource
        super().__init__(code="DuplicateField", message=message)

    def __repr__(self) -> str:
        return (
            "DuplicateFieldError(code=%r, message=%r, field=%r, duplicate_value=%r, conflicting_resource=%r)"
            % (
                self.code,
                self.message,
                self.field,
                self.duplicate_value,
                self.conflicting_resource,
            )
        )


class DuplicateFieldWithConflictingResourceError(ErrorObject):
    #: :class:`str`
    field: str
    #: :class:`typing.Any` `(Named` ``duplicateValue`` `in Commercetools)`
    duplicate_value: typing.Any
    #: :class:`commercetools.types.Reference` `(Named` ``conflictingResource`` `in Commercetools)`
    conflicting_resource: "Reference"

    def __init__(
        self,
        *,
        message: str,
        field: str,
        duplicate_value: typing.Any,
        conflicting_resource: "Reference"
    ) -> None:
        self.field = field
        self.duplicate_value = duplicate_value
        self.conflicting_resource = conflicting_resource
        super().__init__(code="DuplicateFieldWithConflictingResource", message=message)

    def __repr__(self) -> str:
        return (
            "DuplicateFieldWithConflictingResourceError(code=%r, message=%r, field=%r, duplicate_value=%r, conflicting_resource=%r)"
            % (
                self.code,
                self.message,
                self.field,
                self.duplicate_value,
                self.conflicting_resource,
            )
        )


class DuplicatePriceScopeError(ErrorObject):
    #: List of :class:`commercetools.types.Price` `(Named` ``conflictingPrices`` `in Commercetools)`
    conflicting_prices: typing.List["Price"]

    def __init__(
        self, *, message: str, conflicting_prices: typing.List["Price"]
    ) -> None:
        self.conflicting_prices = conflicting_prices
        super().__init__(code="DuplicatePriceScope", message=message)

    def __repr__(self) -> str:
        return (
            "DuplicatePriceScopeError(code=%r, message=%r, conflicting_prices=%r)"
            % (self.code, self.message, self.conflicting_prices)
        )


class DuplicateVariantValuesError(ErrorObject):
    #: :class:`commercetools.types.VariantValues` `(Named` ``variantValues`` `in Commercetools)`
    variant_values: "VariantValues"

    def __init__(self, *, message: str, variant_values: "VariantValues") -> None:
        self.variant_values = variant_values
        super().__init__(code="DuplicateVariantValues", message=message)

    def __repr__(self) -> str:
        return "DuplicateVariantValuesError(code=%r, message=%r, variant_values=%r)" % (
            self.code,
            self.message,
            self.variant_values,
        )


class EnumValueIsUsedError(ErrorObject):
    def __init__(self, *, message: str) -> None:
        super().__init__(code="EnumValueIsUsed", message=message)

    def __repr__(self) -> str:
        return "EnumValueIsUsedError(code=%r, message=%r)" % (self.code, self.message)


class ExtensionBadResponseError(ErrorObject):
    #: Optional :class:`commercetools.types.LocalizedString` `(Named` ``localizedMessage`` `in Commercetools)`
    localized_message: typing.Optional["LocalizedString"]
    #: Optional :class:`object` `(Named` ``extensionExtraInfo`` `in Commercetools)`
    extension_extra_info: typing.Optional[object]
    #: :class:`commercetools.types.ErrorByExtension` `(Named` ``errorByExtension`` `in Commercetools)`
    error_by_extension: "ErrorByExtension"

    def __init__(
        self,
        *,
        message: str,
        error_by_extension: "ErrorByExtension",
        localized_message: typing.Optional["LocalizedString"] = None,
        extension_extra_info: typing.Optional[object] = None
    ) -> None:
        self.localized_message = localized_message
        self.extension_extra_info = extension_extra_info
        self.error_by_extension = error_by_extension
        super().__init__(code="ExtensionBadResponse", message=message)

    def __repr__(self) -> str:
        return (
            "ExtensionBadResponseError(code=%r, message=%r, localized_message=%r, extension_extra_info=%r, error_by_extension=%r)"
            % (
                self.code,
                self.message,
                self.localized_message,
                self.extension_extra_info,
                self.error_by_extension,
            )
        )


class ExtensionNoResponseError(ErrorObject):
    #: Optional :class:`commercetools.types.LocalizedString` `(Named` ``localizedMessage`` `in Commercetools)`
    localized_message: typing.Optional["LocalizedString"]
    #: Optional :class:`object` `(Named` ``extensionExtraInfo`` `in Commercetools)`
    extension_extra_info: typing.Optional[object]
    #: :class:`commercetools.types.ErrorByExtension` `(Named` ``errorByExtension`` `in Commercetools)`
    error_by_extension: "ErrorByExtension"

    def __init__(
        self,
        *,
        message: str,
        error_by_extension: "ErrorByExtension",
        localized_message: typing.Optional["LocalizedString"] = None,
        extension_extra_info: typing.Optional[object] = None
    ) -> None:
        self.localized_message = localized_message
        self.extension_extra_info = extension_extra_info
        self.error_by_extension = error_by_extension
        super().__init__(code="ExtensionNoResponse", message=message)

    def __repr__(self) -> str:
        return (
            "ExtensionNoResponseError(code=%r, message=%r, localized_message=%r, extension_extra_info=%r, error_by_extension=%r)"
            % (
                self.code,
                self.message,
                self.localized_message,
                self.extension_extra_info,
                self.error_by_extension,
            )
        )


class ExtensionUpdateActionsFailedError(ErrorObject):
    #: Optional :class:`commercetools.types.LocalizedString` `(Named` ``localizedMessage`` `in Commercetools)`
    localized_message: typing.Optional["LocalizedString"]
    #: Optional :class:`object` `(Named` ``extensionExtraInfo`` `in Commercetools)`
    extension_extra_info: typing.Optional[object]
    #: :class:`commercetools.types.ErrorByExtension` `(Named` ``errorByExtension`` `in Commercetools)`
    error_by_extension: "ErrorByExtension"

    def __init__(
        self,
        *,
        message: str,
        error_by_extension: "ErrorByExtension",
        localized_message: typing.Optional["LocalizedString"] = None,
        extension_extra_info: typing.Optional[object] = None
    ) -> None:
        self.localized_message = localized_message
        self.extension_extra_info = extension_extra_info
        self.error_by_extension = error_by_extension
        super().__init__(code="ExtensionUpdateActionsFailed", message=message)

    def __repr__(self) -> str:
        return (
            "ExtensionUpdateActionsFailedError(code=%r, message=%r, localized_message=%r, extension_extra_info=%r, error_by_extension=%r)"
            % (
                self.code,
                self.message,
                self.localized_message,
                self.extension_extra_info,
                self.error_by_extension,
            )
        )


class InsufficientScopeError(ErrorObject):
    def __init__(self, *, message: str) -> None:
        super().__init__(code="insufficient_scope", message=message)

    def __repr__(self) -> str:
        return "InsufficientScopeError(code=%r, message=%r)" % (self.code, self.message)


class InvalidCredentialsError(ErrorObject):
    def __init__(self, *, message: str) -> None:
        super().__init__(code="InvalidCredentials", message=message)

    def __repr__(self) -> str:
        return "InvalidCredentialsError(code=%r, message=%r)" % (
            self.code,
            self.message,
        )


class InvalidCurrentPasswordError(ErrorObject):
    def __init__(self, *, message: str) -> None:
        super().__init__(code="InvalidCurrentPassword", message=message)

    def __repr__(self) -> str:
        return "InvalidCurrentPasswordError(code=%r, message=%r)" % (
            self.code,
            self.message,
        )


class InvalidFieldError(ErrorObject):
    #: :class:`str`
    field: str
    #: :class:`typing.Any` `(Named` ``invalidValue`` `in Commercetools)`
    invalid_value: typing.Any
    #: Optional :class:`list` `(Named` ``allowedValues`` `in Commercetools)`
    allowed_values: typing.Optional[list]

    def __init__(
        self,
        *,
        message: str,
        field: str,
        invalid_value: typing.Any,
        allowed_values: typing.Optional[list] = None
    ) -> None:
        self.field = field
        self.invalid_value = invalid_value
        self.allowed_values = allowed_values
        super().__init__(code="InvalidField", message=message)

    def __repr__(self) -> str:
        return (
            "InvalidFieldError(code=%r, message=%r, field=%r, invalid_value=%r, allowed_values=%r)"
            % (
                self.code,
                self.message,
                self.field,
                self.invalid_value,
                self.allowed_values,
            )
        )


class InvalidInputError(ErrorObject):
    def __init__(self, *, message: str) -> None:
        super().__init__(code="InvalidInput", message=message)

    def __repr__(self) -> str:
        return "InvalidInputError(code=%r, message=%r)" % (self.code, self.message)


class InvalidItemShippingDetailsError(ErrorObject):
    #: :class:`str`
    subject: str
    #: :class:`str` `(Named` ``itemId`` `in Commercetools)`
    item_id: str

    def __init__(self, *, message: str, subject: str, item_id: str) -> None:
        self.subject = subject
        self.item_id = item_id
        super().__init__(code="InvalidItemShippingDetails", message=message)

    def __repr__(self) -> str:
        return (
            "InvalidItemShippingDetailsError(code=%r, message=%r, subject=%r, item_id=%r)"
            % (self.code, self.message, self.subject, self.item_id)
        )


class InvalidJsonInputError(ErrorObject):
    def __init__(self, *, message: str) -> None:
        super().__init__(code="InvalidJsonInput", message=message)

    def __repr__(self) -> str:
        return "InvalidJsonInputError(code=%r, message=%r)" % (self.code, self.message)


class InvalidOperationError(ErrorObject):
    def __init__(self, *, message: str) -> None:
        super().__init__(code="InvalidOperation", message=message)

    def __repr__(self) -> str:
        return "InvalidOperationError(code=%r, message=%r)" % (self.code, self.message)


class InvalidSubjectError(ErrorObject):
    def __init__(self, *, message: str) -> None:
        super().__init__(code="InvalidSubject", message=message)

    def __repr__(self) -> str:
        return "InvalidSubjectError(code=%r, message=%r)" % (self.code, self.message)


class InvalidTokenError(ErrorObject):
    def __init__(self, *, message: str) -> None:
        super().__init__(code="invalid_token", message=message)

    def __repr__(self) -> str:
        return "InvalidTokenError(code=%r, message=%r)" % (self.code, self.message)


class LanguageUsedInStoresError(ErrorObject):
    def __init__(self, *, message: str) -> None:
        super().__init__(code="LanguageUsedInStores", message=message)

    def __repr__(self) -> str:
        return "LanguageUsedInStoresError(code=%r, message=%r)" % (
            self.code,
            self.message,
        )


class MatchingPriceNotFoundError(ErrorObject):
    #: :class:`str` `(Named` ``productId`` `in Commercetools)`
    product_id: str
    #: :class:`int` `(Named` ``variantId`` `in Commercetools)`
    variant_id: int
    #: Optional :class:`str`
    currency: typing.Optional[str]
    #: Optional :class:`str`
    country: typing.Optional[str]
    #: Optional :class:`commercetools.types.CustomerGroupReference` `(Named` ``customerGroup`` `in Commercetools)`
    customer_group: typing.Optional["CustomerGroupReference"]
    #: Optional :class:`commercetools.types.ChannelReference`
    channel: typing.Optional["ChannelReference"]

    def __init__(
        self,
        *,
        message: str,
        product_id: str,
        variant_id: int,
        currency: typing.Optional[str] = None,
        country: typing.Optional[str] = None,
        customer_group: typing.Optional["CustomerGroupReference"] = None,
        channel: typing.Optional["ChannelReference"] = None
    ) -> None:
        self.product_id = product_id
        self.variant_id = variant_id
        self.currency = currency
        self.country = country
        self.customer_group = customer_group
        self.channel = channel
        super().__init__(code="MatchingPriceNotFound", message=message)

    def __repr__(self) -> str:
        return (
            "MatchingPriceNotFoundError(code=%r, message=%r, product_id=%r, variant_id=%r, currency=%r, country=%r, customer_group=%r, channel=%r)"
            % (
                self.code,
                self.message,
                self.product_id,
                self.variant_id,
                self.currency,
                self.country,
                self.customer_group,
                self.channel,
            )
        )


class MissingTaxRateForCountryError(ErrorObject):
    #: :class:`str` `(Named` ``taxCategoryId`` `in Commercetools)`
    tax_category_id: str
    #: Optional :class:`str`
    country: typing.Optional[str]
    #: Optional :class:`str`
    state: typing.Optional[str]

    def __init__(
        self,
        *,
        message: str,
        tax_category_id: str,
        country: typing.Optional[str] = None,
        state: typing.Optional[str] = None
    ) -> None:
        self.tax_category_id = tax_category_id
        self.country = country
        self.state = state
        super().__init__(code="MissingTaxRateForCountry", message=message)

    def __repr__(self) -> str:
        return (
            "MissingTaxRateForCountryError(code=%r, message=%r, tax_category_id=%r, country=%r, state=%r)"
            % (self.code, self.message, self.tax_category_id, self.country, self.state)
        )


class NoMatchingProductDiscountFoundError(ErrorObject):
    def __init__(self, *, message: str) -> None:
        super().__init__(code="NoMatchingProductDiscountFound", message=message)

    def __repr__(self) -> str:
        return "NoMatchingProductDiscountFoundError(code=%r, message=%r)" % (
            self.code,
            self.message,
        )


class OutOfStockError(ErrorObject):
    #: List of :class:`str` `(Named` ``lineItems`` `in Commercetools)`
    line_items: typing.List[str]
    #: List of :class:`str`
    skus: typing.List[str]

    def __init__(
        self, *, message: str, line_items: typing.List[str], skus: typing.List[str]
    ) -> None:
        self.line_items = line_items
        self.skus = skus
        super().__init__(code="OutOfStock", message=message)

    def __repr__(self) -> str:
        return "OutOfStockError(code=%r, message=%r, line_items=%r, skus=%r)" % (
            self.code,
            self.message,
            self.line_items,
            self.skus,
        )


class PriceChangedError(ErrorObject):
    #: List of :class:`str` `(Named` ``lineItems`` `in Commercetools)`
    line_items: typing.List[str]
    #: :class:`bool`
    shipping: bool

    def __init__(
        self, *, message: str, line_items: typing.List[str], shipping: bool
    ) -> None:
        self.line_items = line_items
        self.shipping = shipping
        super().__init__(code="PriceChanged", message=message)

    def __repr__(self) -> str:
        return "PriceChangedError(code=%r, message=%r, line_items=%r, shipping=%r)" % (
            self.code,
            self.message,
            self.line_items,
            self.shipping,
        )


class ReferenceExistsError(ErrorObject):
    #: Optional :class:`commercetools.types.ReferenceTypeId` `(Named` ``referencedBy`` `in Commercetools)`
    referenced_by: typing.Optional["ReferenceTypeId"]

    def __init__(
        self, *, message: str, referenced_by: typing.Optional["ReferenceTypeId"] = None
    ) -> None:
        self.referenced_by = referenced_by
        super().__init__(code="ReferenceExists", message=message)

    def __repr__(self) -> str:
        return "ReferenceExistsError(code=%r, message=%r, referenced_by=%r)" % (
            self.code,
            self.message,
            self.referenced_by,
        )


class ReferencedResourceNotFoundError(ErrorObject):
    #: :class:`commercetools.types.ReferenceTypeId` `(Named` ``typeId`` `in Commercetools)`
    type_id: "ReferenceTypeId"
    #: Optional :class:`str`
    id: typing.Optional[str]
    #: Optional :class:`str`
    key: typing.Optional[str]

    def __init__(
        self,
        *,
        message: str,
        type_id: "ReferenceTypeId",
        id: typing.Optional[str] = None,
        key: typing.Optional[str] = None
    ) -> None:
        self.type_id = type_id
        self.id = id
        self.key = key
        super().__init__(code="ReferencedResourceNotFound", message=message)

    def __repr__(self) -> str:
        return (
            "ReferencedResourceNotFoundError(code=%r, message=%r, type_id=%r, id=%r, key=%r)"
            % (self.code, self.message, self.type_id, self.id, self.key)
        )


class RequiredFieldError(ErrorObject):
    #: :class:`str`
    field: str

    def __init__(self, *, message: str, field: str) -> None:
        self.field = field
        super().__init__(code="RequiredField", message=message)

    def __repr__(self) -> str:
        return "RequiredFieldError(code=%r, message=%r, field=%r)" % (
            self.code,
            self.message,
            self.field,
        )


class ResourceNotFoundError(ErrorObject):
    def __init__(self, *, message: str) -> None:
        super().__init__(code="ResourceNotFound", message=message)

    def __repr__(self) -> str:
        return "ResourceNotFoundError(code=%r, message=%r)" % (self.code, self.message)


class ShippingMethodDoesNotMatchCartError(ErrorObject):
    def __init__(self, *, message: str) -> None:
        super().__init__(code="ShippingMethodDoesNotMatchCart", message=message)

    def __repr__(self) -> str:
        return "ShippingMethodDoesNotMatchCartError(code=%r, message=%r)" % (
            self.code,
            self.message,
        )



import datetime

from commercetools import types


def _get_draft_object(**kwargs):
    customer = {
        "first_name": "John",
        "last_name": "Doe",
        "title": "mr",
        "date_of_birth": datetime.date(year=1985, month=8, day=4),
        "email": "info@example.org",
        "password": "notsosecret",
    }
    customer.update(kwargs)
    return types.CustomerDraft(**customer)


def test_get_by_id(client):
    result = client.customers.create(draft=_get_draft_object())
    assert result.customer.id

    customer = client.customers.get_by_id(result.customer.id)
    assert customer


def test_get_by_key(client):
    result = client.customers.create(draft=_get_draft_object(key="test-customer"))
    assert result.customer.key
    assert result.cart is None

    customer = client.customers.get_by_key(result.customer.key)
    assert customer


def test_query(client):
    client.customers.create(draft=_get_draft_object(key="test-customer-1"))
    client.customers.create(draft=_get_draft_object(key="test-customer-2"))

    result = client.customers.query(sort="id asc", limit=10)
    assert len(result.results) == 2
    assert result.total == 2

    result = client.customers.query(sort=["id asc", "name asc"], limit=1)
    assert len(result.results) == 1
    assert result.total == 2


def test_delete_by_id(client):
    result = client.customers.create(draft=_get_draft_object())
    assert result.customer.id
    assert client.customers.delete_by_id(
        result.customer.id, version=result.customer.version
    )


def test_delete_by_key(client):
    result = client.customers.create(draft=_get_draft_object(key="test-customer"))
    assert result.customer.key
    assert client.customers.delete_by_key(
        result.customer.key, version=result.customer.version
    )



# DO NOT EDIT! This file is automatically generated
import datetime
import typing

from commercetools.types._abstract import _BaseType

__all__ = ["ApiClient", "ApiClientDraft", "ApiClientPagedQueryResponse"]


class ApiClient(_BaseType):
    #: :class:`str`
    id: str
    #: :class:`str`
    name: str
    #: :class:`str`
    scope: str
    #: Optional :class:`datetime.datetime` `(Named` ``createdAt`` `in Commercetools)`
    created_at: typing.Optional[datetime.datetime]
    #: Optional :class:`datetime.date` `(Named` ``lastUsedAt`` `in Commercetools)`
    last_used_at: typing.Optional[datetime.date]
    #: Optional :class:`datetime.datetime` `(Named` ``deleteAt`` `in Commercetools)`
    delete_at: typing.Optional[datetime.datetime]
    #: Optional :class:`str`
    secret: typing.Optional[str]

    def __init__(
        self,
        *,
        id: str,
        name: str,
        scope: str,
        created_at: typing.Optional[datetime.datetime] = None,
        last_used_at: typing.Optional[datetime.date] = None,
        delete_at: typing.Optional[datetime.datetime] = None,
        secret: typing.Optional[str] = None
    ) -> None:
        self.id = id
        self.name = name
        self.scope = scope
        self.created_at = created_at
        self.last_used_at = last_used_at
        self.delete_at = delete_at
        self.secret = secret
        super().__init__()

    def __repr__(self) -> str:
        return (
            "ApiClient(id=%r, name=%r, scope=%r, created_at=%r, last_used_at=%r, delete_at=%r, secret=%r)"
            % (
                self.id,
                self.name,
                self.scope,
                self.created_at,
                self.last_used_at,
                self.delete_at,
                self.secret,
            )
        )


class ApiClientDraft(_BaseType):
    #: :class:`str`
    name: str
    #: :class:`str`
    scope: str
    #: Optional :class:`int` `(Named` ``deleteDaysAfterCreation`` `in Commercetools)`
    delete_days_after_creation: typing.Optional[int]

    def __init__(
        self,
        *,
        name: str,
        scope: str,
        delete_days_after_creation: typing.Optional[int] = None
    ) -> None:
        self.name = name
        self.scope = scope
        self.delete_days_after_creation = delete_days_after_creation
        super().__init__()

    def __repr__(self) -> str:
        return "ApiClientDraft(name=%r, scope=%r, delete_days_after_creation=%r)" % (
            self.name,
            self.scope,
            self.delete_days_after_creation,
        )


class ApiClientPagedQueryResponse(_BaseType):
    #: :class:`int`
    limit: int
    #: :class:`int`
    count: int
    #: Optional :class:`int`
    total: typing.Optional[int]
    #: :class:`int`
    offset: int
    #: List of :class:`commercetools.types.ApiClient`
    results: typing.Sequence["ApiClient"]

    def __init__(
        self,
        *,
        limit: int,
        count: int,
        offset: int,
        results: typing.Sequence["ApiClient"],
        total: typing.Optional[int] = None
    ) -> None:
        self.limit = limit
        self.count = count
        self.total = total
        self.offset = offset
        self.results = results
        super().__init__()

    def __repr__(self) -> str:
        return (
            "ApiClientPagedQueryResponse(limit=%r, count=%r, total=%r, offset=%r, results=%r)"
            % (self.limit, self.count, self.total, self.offset, self.results)
        )



import threading
import time
import typing

import pytest
import requests

from commercetools import Client
from commercetools.testing import backend_mocker
from commercetools.testing.server import Server


@pytest.fixture(autouse=True)
def reset_token_cache():
    from commercetools.utils import DefaultTokenSaver

    DefaultTokenSaver.clear_cache()


@pytest.fixture()
def commercetools_api():
    with backend_mocker() as m:
        yield m


@pytest.fixture
def commercetools_client(commercetools_api) -> typing.Generator[Client, None, None]:
    yield Client(
        project_key="unittest",
        client_id="client-id",
        client_secret="client-secret",
        scope=[],
        url="https://api.sphere.io",
        token_url="https://auth.sphere.io/oauth/token",
    )


@pytest.fixture()
def commercetools_http_server(commercetools_api):
    is_running = threading.Event()
    server = Server(commercetools_api)

    def serve():
        from werkzeug.serving import run_simple

        is_running.set()
        server.api_url = "http://localhost:8989"
        run_simple("localhost", port=8989, application=server)

    thread = threading.Thread(target=serve, daemon=True)
    thread.start()

    if is_running.wait():
        for i in range(0, 5):
            response = requests.get(server.api_url + "/-/health")
            if response.status_code == 200:
                break
            time.sleep(0.5)

        yield server

    thread.join(timeout=0)



# DO NOT EDIT! This file is automatically generated
import typing

import marshmallow
from marshmallow import fields

from commercetools._schemas._product import (
    ProductProjectionPagedQueryResponseSchema,
    ProductProjectionPagedSearchResponseSchema,
    ProductProjectionSchema,
)
from commercetools.helpers import OptionalList, RemoveEmptyValuesMixin
from commercetools.types._product import (
    ProductProjection,
    ProductProjectionPagedQueryResponse,
    ProductProjectionPagedSearchResponse,
)
from commercetools.typing import OptionalListStr

from . import abstract, traits


class _ProductProjectionGetSchema(traits.ExpandableSchema, traits.PriceSelectingSchema):
    staged = fields.Bool(required=False, missing=False)


class _ProductProjectionQuerySchema(
    traits.ExpandableSchema,
    traits.SortableSchema,
    traits.PagingSchema,
    traits.QuerySchema,
    traits.PriceSelectingSchema,
):
    staged = fields.Bool(required=False, missing=False)


class _ProductProjectionSearchSchema(
    traits.SortableSchema,
    traits.PagingSchema,
    traits.PriceSelectingSchema,
    traits.ExpandableSchema,
):
    fuzzy = fields.Bool(required=False, missing=False)
    fuzzy_level = fields.Int(data_key="fuzzyLevel", required=False)
    mark_matching_variants = fields.Bool(data_key="markMatchingVariants", missing=False)
    staged = fields.Bool(required=False, missing=False)
    filter = OptionalList(fields.String(), required=False)
    filter_facets = OptionalList(
        fields.String(), data_key="filter.facets", required=False
    )
    filter_query = OptionalList(
        fields.String(), data_key="filter.query", required=False
    )
    facet = OptionalList(fields.String(), required=False)
    text = fields.Dict()

    @marshmallow.post_dump
    def _text_post_dump(self, data, **kwrags):
        values = data.pop("text")
        if not values:
            return data
        for key, val in values.items():
            data[f"text.{key}"] = val
        return data

    @marshmallow.pre_load
    def _text_post_load(self, data, **kwrags):
        items = {}
        for key in list(data.keys()):
            if key.startswith("text."):
                items[key[5:]] = data[key]
                del data[key]
        data["text"] = items
        return data


class _ProductProjectionSuggestSchema(traits.SortableSchema, traits.PagingSchema):
    fuzzy = fields.Bool(required=False, missing=False)
    staged = fields.Bool(required=False, missing=False)
    search_keywords = fields.Dict()

    @marshmallow.post_dump
    def _search_keywords_post_dump(self, data, **kwrags):
        values = data.pop("search_keywords")
        if not values:
            return data
        for key, val in values.items():
            data[f"searchKeywords.{key}"] = val
        return data

    @marshmallow.pre_load
    def _search_keywords_post_load(self, data, **kwrags):
        items = {}
        for key in list(data.keys()):
            if key.startswith("searchKeywords."):
                items[key[15:]] = data[key]
                del data[key]
        data["search_keywords"] = items
        return data


class ProductProjectionService(abstract.AbstractService):
    """A projected representation of a product shows the product with its current
    or staged data.

    The current or staged representation of a product in a catalog is called a
    product projection.
    """

    def get_by_id(
        self,
        id: str,
        *,
        expand: OptionalListStr = None,
        price_currency: OptionalListStr = None,
        price_country: OptionalListStr = None,
        price_customer_group: OptionalListStr = None,
        price_channel: OptionalListStr = None,
        locale_projection: OptionalListStr = None,
        store_projection: OptionalListStr = None,
        staged: bool = None,
    ) -> ProductProjection:
        """Gets the current or staged representation of a product in a catalog by
        ID.

        When used with an API client that has the
        view_published_products:{projectKey} scope, this endpoint only returns
        published (current) product projections.
        """
        params = self._serialize_params(
            {
                "expand": expand,
                "price_currency": price_currency,
                "price_country": price_country,
                "price_customer_group": price_customer_group,
                "price_channel": price_channel,
                "locale_projection": locale_projection,
                "store_projection": store_projection,
                "staged": staged,
            },
            _ProductProjectionGetSchema,
        )
        return self._client._get(
            endpoint=f"product-projections/{id}",
            params=params,
            schema_cls=ProductProjectionSchema,
        )

    def get_by_key(
        self,
        key: str,
        *,
        expand: OptionalListStr = None,
        price_currency: OptionalListStr = None,
        price_country: OptionalListStr = None,
        price_customer_group: OptionalListStr = None,
        price_channel: OptionalListStr = None,
        locale_projection: OptionalListStr = None,
        store_projection: OptionalListStr = None,
        staged: bool = None,
    ) -> ProductProjection:
        """Gets the current or staged representation of a product found by Key.

        When used with an API client that has the
        view_published_products:{projectKey} scope, this endpoint only returns
        published (current) product projections.
        """
        params = self._serialize_params(
            {
                "expand": expand,
                "price_currency": price_currency,
                "price_country": price_country,
                "price_customer_group": price_customer_group,
                "price_channel": price_channel,
                "locale_projection": locale_projection,
                "store_projection": store_projection,
                "staged": staged,
            },
            _ProductProjectionGetSchema,
        )
        return self._client._get(
            endpoint=f"product-projections/key={key}",
            params=params,
            schema_cls=ProductProjectionSchema,
        )

    def query(
        self,
        *,
        expand: OptionalListStr = None,
        sort: OptionalListStr = None,
        limit: int = None,
        offset: int = None,
        with_total: bool = None,
        where: OptionalListStr = None,
        predicate_var: typing.Dict[str, str] = None,
        price_currency: OptionalListStr = None,
        price_country: OptionalListStr = None,
        price_customer_group: OptionalListStr = None,
        price_channel: OptionalListStr = None,
        locale_projection: OptionalListStr = None,
        store_projection: OptionalListStr = None,
        staged: bool = None,
    ) -> ProductProjectionPagedQueryResponse:
        """You can use the product projections query endpoint to get the current
        or staged representations of Products.

        When used with an API client that has the
        view_published_products:{projectKey} scope, this endpoint only returns
        published (current) product projections.   A projected representation of
        a product shows the product with its current or staged data. The current
        or staged representation of a product in a catalog is called a product
        projection.
        """
        params = self._serialize_params(
            {
                "expand": expand,
                "sort": sort,
                "limit": limit,
                "offset": offset,
                "with_total": with_total,
                "where": where,
                "predicate_var": predicate_var,
                "price_currency": price_currency,
                "price_country": price_country,
                "price_customer_group": price_customer_group,
                "price_channel": price_channel,
                "locale_projection": locale_projection,
                "store_projection": store_projection,
                "staged": staged,
            },
            _ProductProjectionQuerySchema,
        )
        return self._client._get(
            endpoint="product-projections",
            params=params,
            schema_cls=ProductProjectionPagedQueryResponseSchema,
        )

    def search(
        self,
        mark_matching_variants: bool,
        *,
        sort: OptionalListStr = None,
        limit: int = None,
        offset: int = None,
        with_total: bool = None,
        price_currency: OptionalListStr = None,
        price_country: OptionalListStr = None,
        price_customer_group: OptionalListStr = None,
        price_channel: OptionalListStr = None,
        locale_projection: OptionalListStr = None,
        store_projection: OptionalListStr = None,
        expand: OptionalListStr = None,
        fuzzy: bool = None,
        fuzzy_level: int = None,
        staged: bool = None,
        filter: str = None,
        filter_facets: str = None,
        filter_query: str = None,
        facet: str = None,
        text: typing.Dict[str, str] = None,
    ) -> ProductProjectionPagedSearchResponse:
        """Search Product Projection

        This endpoint provides high performance search queries over
        ProductProjections. The query result contains the ProductProjections for
        which at least one ProductVariant matches the search query. This means
        that variants can be included in the result also for which the search
        query does not match. To determine which ProductVariants match the search
        query, the returned ProductProjections include the additional field
        isMatchingVariant.
        """
        params = self._serialize_params(
            {
                "sort": sort,
                "limit": limit,
                "offset": offset,
                "with_total": with_total,
                "price_currency": price_currency,
                "price_country": price_country,
                "price_customer_group": price_customer_group,
                "price_channel": price_channel,
                "locale_projection": locale_projection,
                "store_projection": store_projection,
                "expand": expand,
                "fuzzy": fuzzy,
                "fuzzy_level": fuzzy_level,
                "mark_matching_variants": mark_matching_variants,
                "staged": staged,
                "filter": filter,
                "filter_facets": filter_facets,
                "filter_query": filter_query,
                "facet": facet,
                "text": text,
            },
            _ProductProjectionSearchSchema,
        )
        return self._client._get(
            endpoint="product-projections/search",
            params=params,
            schema_cls=ProductProjectionPagedSearchResponseSchema,
        )

    def suggest(
        self,
        *,
        sort: OptionalListStr = None,
        limit: int = None,
        offset: int = None,
        with_total: bool = None,
        fuzzy: bool = None,
        staged: bool = None,
        search_keywords: typing.Dict[str, str] = None,
    ) -> ProductProjection:
        """The source of data for suggestions is the searchKeyword field in a
        product
        """
        params = self._serialize_params(
            {
                "sort": sort,
                "limit": limit,
                "offset": offset,
                "with_total": with_total,
                "fuzzy": fuzzy,
                "staged": staged,
                "search_keywords": search_keywords,
            },
            _ProductProjectionSuggestSchema,
        )
        return self._client._get(
            endpoint="product-projections/suggest",
            params=params,
            schema_cls=ProductProjectionSchema,
        )
